/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "cidev.h"
#include "libparser.h"


void free_config(cidev_entry_t *entry)
{
  cidev_entry_t *cur, *ent=entry;
  
  while(ent) 
  {
    cur=ent->next;
    free(ent);
    ent=cur;
  }
}

int read_config_file(char *config, cidev_entry_t **entry)
{
  FILE *infile;
  unsigned char buf[CONFIG_LINE_LEN];
  unsigned char *bufp, *colon, *p, *cur_ent_add = NULL;
  cidev_entry_t *ent = NULL, *cur_ent = NULL;
  int get_kw = -1;
  int kws=0;

  if (!config)
    infile = stdin;
  else
  {
    if (!(infile = fopen(config, "r")))
    {
       return -1;
    }
  }
    
  while (!feof(infile))
  {
    memset(buf, 0, sizeof(buf));
    if (!fgets(buf, sizeof(buf), infile))
    break;
    bufp = buf;


    /* skip whitespace at start of line */
    while (*bufp == ' ' || *bufp == '\t')
      bufp++;
    
    /* ignore comments and empty lines */
    if (*bufp == '#' || *bufp == '\n')
      continue;

    /* a line starting with a string of alphanumeric characters followed
     * by a colon is the keyword of a new entry */
    /* decide if this line could be a keyword */
    colon = strchr(bufp, ':');
    if (colon && bufp == buf)
    {
      int isa_kw = 1;
      for (p = bufp; p != colon; p++)
      {
        if (!isalnum(*p))
        {
          isa_kw = 0;
          break;
        }
      }

      if (isa_kw) 
        get_kw = 1;
    }

    if (get_kw > 0)
    {
      kws++;

      /* create new entry */

      ent = (cidev_entry_t *)malloc(sizeof(cidev_entry_t));
      if (!ent)
        return -ENOMEM;

      memset(ent, 0, sizeof(cidev_entry_t));

      if (!cur_ent)
        *entry = ent;
      else
        cur_ent->next = ent;

      cur_ent = ent;
      cur_ent_add = cur_ent->string;

      *colon = '\0';
      strncpy(cur_ent->keyword, bufp, CIDEV_KEYWORD_SLEN);
      bufp = colon+1;

      get_kw = 0;

      /* ignore trailing whitespace */
      while (*bufp == ' ' || *bufp == '\t')
        bufp++;

      /* ignore trailing comments */
      if (*bufp == '#' || *bufp == '\n' || *bufp == '\0')
        continue;

      /* parse error... extra junk trailing `keyword:' */
      free_config(*entry);
      entry = NULL;
      return -1;
    }

    /* if this isn't a keyword line, then just take everything
     * up until '\n' as the field */
    if (ent)
    {
      int valid=FALSE, ws=FALSE, pos=0;
      while ((*bufp != '#')  &&
             (*bufp != '\n') &&
             (cur_ent_add < (cur_ent->string + CIDEV_STRING_SLEN)))
      {   
        if (!valid)
        {   
          if (!pos && *bufp == '=')
            break;

          if (*bufp == ':')
            break;

          if(ws)
          {   
            if (*bufp == '=')
              valid = TRUE;
            else if(*bufp != '\t' && *bufp != ' ')
              break;
          }
          else
          {   
            if (*bufp == '=')
              valid = TRUE;
            else if(*bufp == '\t' || *bufp == ' ')
              ws = TRUE;
          }
        }

        *cur_ent_add++ = *bufp++;
        pos++;
      }

      if(valid)
        *cur_ent_add++ = '\n';
      else
      {   
        free_config(*entry);
        entry = NULL;
        return -1;
      }
    }
    else
    {
      return -1;
    }

  }
  
  fclose(infile);
  return (get_kw < 0) ? -1 : 0 ;
}

int print_config_file(cidev_entry_t *config)
{
  cidev_entry_t *cur_ent;
  for(cur_ent=config;cur_ent;cur_ent = cur_ent->next)
  {
     printf("%s:\n%s\n", cur_ent->keyword, cur_ent->string);
  }
  return 0;
}


/**
 * text_get_entry_keyword
 *
 * Reads n'th entry with keyword.  First entry is n = 1
 */
cidev_entry_t *text_get_entry_keyword(cidev_entry_t *config, char *kw, int n)
{
  int pos=1;
  cidev_entry_t *entry;
  
  if (n<-1)
     return NULL;

  for( entry = config; entry; entry = entry->next )
  {
    if (!strcmp(entry->keyword,kw))
    {
      if (n==pos)
        break;
      else
        pos++;
    }
  }

  return entry;
}


/**
 * get_stomith_text_entry
 * @config: cidev 
 * @keyword: entry name
 * @key: entry unique identifier
 * @value: entry to fetch
 *
 * Returns: NULL or cidev entry
 */

cidev_entry_t *text_get_entry(cidev_entry_t *config, char *kw, char *key, char *value)
{
  cidev_entry_t *entry;
  unsigned char *str, valuebuf[128];
  int i, len;

  if(kw==NULL || key==NULL)
    return NULL;

  for (i = 1; ; i++)
  {
    entry = text_get_entry_keyword(config, kw, i);
    if (entry==NULL)
      break;

    str = get_string_val(entry->string, key, &len);
    if (!str)
      continue;

    str = strip_white_space(str,&len);

    memset(valuebuf, 0, 128);
    strncpy(valuebuf, str, len);

    if (strlen(value) != strlen(valuebuf) ||
        memcmp(str, value, strlen(value)))
      continue;

    return entry;
  }

  return NULL;
}


/**
 * strip_white_space
 * @str: string to scan
 * @len: integer for length of stripped string
 *
 * Returns: pointer to the first non-whitespace char
 *
 * FIXME This should go into a differnt utility file 
 */
char *strip_white_space(char *str,int *len)
{
  char *s;

  if (!str)
    return NULL;

  *len = strlen(str);

  /* find first newline */
  s=strchr(str,'\n');
  if (s)
  {
    *len=(s-str);
    s--;
  }
  else
    s=(*len+str-1);

  /* strip trailing whitespace */
  for( ; s != str ; s--)
  {
    if ((*s == ' ' ) || (*s == '\t') || (*s == '\n'))
      *len = *len - 1;
    else
      break;
  }

  /* strip leading whitespace */
  for(s=str; *s!='\0'; s++)
  {
    if ((*s == ' ' ) || (*s == '\t'))
      *len = *len - 1;
    else
      break;
  }

  return s;
}

/* vim: set ai cin et sw=2 ts=2 : */
