/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>

#include "osi_list.h"

typedef struct node_fence_s {
  char *nf_name;
  char *nf_key;
  char *nf_param;
  osi_list_t nf_list;
} node_fence_t;

typedef struct node_s {
  char *n_name;
  char *n_ip;
  osi_list_t n_flist; /* node specific fence list */
  osi_list_t n_list;
} node_t;

typedef struct fence_dev_s {
  char *f_name;
  char *f_rname;
  char *f_agent;
  char *f_ip;
  char *f_login;
  char *f_passwd;
  osi_list_t f_list;
} fence_dev_t;

osi_list_decl(fence_list);
osi_list_decl(node_list);

char *fencedev_lookup(char *name){
  osi_list_t *lptr;

  for(lptr = fence_list.next; lptr != &fence_list; lptr = lptr->next){
    fence_dev_t *fence_dev;
    fence_dev = osi_list_entry(lptr, fence_dev_t, f_list);
    if(!strcmp(fence_dev->f_name, name)){
      return fence_dev->f_rname;
    }
  }
  return NULL;
}


char *translate_agent_name(char *agent){
  if(!strcmp(agent, "fence_apc_ms") || !strcmp(agent, "do_apc_ms")){
    return "fence_apc";
  } else if(!strcmp(agent, "fence_brcd_port")||!strcmp(agent, "do_brcd_port")){
    return "fence_brocade";
  } else if(!strcmp(agent, "fence_vxl_zone") || !strcmp(agent, "do_vxl_zone")){
    return "fence_vixel";
  } else if(!strcmp(agent, "fence_wti_nps") || !strcmp(agent, "do_wti_nps")){
    return "fence_wti";
  } else if(!strcmp(agent, "fence_zvm_cp") || !strcmp(agent, "do_zvm_cp")){
    return "fence_zvm";
  } else {
    return agent;
  }
}


int fill_fence(fence_dev_t *fence_dev, FILE *cfg_fp){
  long pos;
  char line[256];
  char param[128];
  int i;
  osi_list_t *lptr;

  pos = ftell(cfg_fp);
  if(pos == -1){
    return -1;
  }

  while(fgets(line, sizeof(line), cfg_fp)){
    if(line[0] == '#'){
      continue;
    } else if(!strncmp(line, "name", 4)){
      sscanf(line, "name = %s", param);
      fence_dev->f_name = strdup(param);
    } else if(!strncmp(line, "agent", 5)){
      sscanf(line, "agent = %s", param);
      fence_dev->f_agent = strdup(translate_agent_name(param));
      fence_dev->f_rname = malloc(strlen(fence_dev->f_agent)+3);
      if(!fence_dev->f_rname || !fence_dev->f_agent){
	fprintf(stderr, "Unable to allocate memory.  Exiting...\n");
	exit(EXIT_FAILURE);
      }
      for(i=1, lptr = fence_list.next; lptr != &fence_list; lptr = lptr->next){
	fence_dev_t *list_dev;
	list_dev = osi_list_entry(lptr, fence_dev_t, f_list);
	if(!strcmp(list_dev->f_agent, fence_dev->f_agent)){
	  i++;
	}
      }
      if(!strncmp(param, "fence_", 6)){
	sprintf(fence_dev->f_rname, "%s_%d", param+6, i);
      } else {
	fence_dev->f_rname = NULL;
      }
      continue;
    } else if(!strncmp(line, "ipaddr", 6)){
      sscanf(line, "ipaddr = %s", param);
      fence_dev->f_ip = strdup(param);
    } else if(!strncmp(line, "login", 5)){
      sscanf(line, "login = %s", param);
      fence_dev->f_login = strdup(param);
    } else if(!strncmp(line, "passwd", 6)){
      sscanf(line, "passwd = %s", param);
      fence_dev->f_passwd = strdup(param);
    } else {
      break;
    }
    pos = ftell(cfg_fp);
    if(pos == -1){
      return -1;
    }
  }

  fseek(cfg_fp, pos, SEEK_SET);
  return 0;
}


int fill_node(node_t *node, FILE *cfg_fp){
  long pos;
  char line[256];
  char fence[128];
  char param[128];
  node_fence_t *nfence;

  pos = ftell(cfg_fp);
  if(pos == -1){
    return -1;
  }

  while(fgets(line, sizeof(line), cfg_fp)){
    if(line[0] == '#'){
      continue;
    } else if(!strncmp(line, "name", 4)){
      sscanf(line, "name = %s", param);
      node->n_name = strdup(param);
    } else if(!strncmp(line, "cid", 3)){
      continue;
    } else if(!strncmp(line, "ipaddr", 6)){
      sscanf(line, "ipaddr = %s", param);
      node->n_ip = strdup(param);

    } else if(!strncmp(line, "fm", 2)){
      sscanf(line, "fm = %s %s", fence, param);
      nfence = malloc(sizeof(node_fence_t));
      memset(nfence, 0, sizeof(node_fence_t));
      nfence->nf_name = strdup(fence);
      nfence->nf_param = strdup(param);
      osi_list_add(&nfence->nf_list, &node->n_flist); 
    } else {
      break;
    }
    pos = ftell(cfg_fp);
    if(pos == -1){
      return -1;
    }
  }

  fseek(cfg_fp, pos, SEEK_SET);
  return 0;
}
    

int generate_files(char *cluster_name){
  FILE *fp;
  osi_list_t *lptr;

  if(mkdir(cluster_name, 00770)){
    fprintf(stderr, "Unable to create directory \"%s\": %s\n",
	    cluster_name, strerror(errno));
    return -1;
  }

  if(chdir(cluster_name)){
    fprintf(stderr, "Unable to enter directory \"%s\": %s\n",
	    cluster_name, strerror(errno));
    return -1;
  }

  fp = fopen("cluster.ccs", "w+");
  if(!fp){
    fprintf(stderr, "Unable to open \"%s/cluster.ccs\": %s\n",
	    cluster_name, strerror(errno));
    return -1;
  }

  fprintf(fp,
	  "cluster {\n"
	  "  name = \"%s\"\n"
	  "  lock_gulm {\n"
	  "# Edit the following line with nodes that will be lock servers\n"
	  "--    servers = [ \"server1\", \"server2\" ]\n"
	  "  }\n"
	  "}\n", cluster_name);
  fclose(fp);

  fp = fopen("fence.ccs", "w+");
  fprintf(fp, "fence_devices {\n");
  
  for(lptr = fence_list.prev; lptr != &fence_list; lptr = lptr->prev){
    fence_dev_t *fence_dev;
    fence_dev = osi_list_entry(lptr, fence_dev_t, f_list);
    fprintf(fp,
	    "  %s {\n"
	    "    agent = \"%s\"\n",
	    (fence_dev->f_rname) ? fence_dev->f_rname:
	    "-- Choose a name for this fence device\n  <choose>",
	    fence_dev->f_agent);
    if(fence_dev->f_ip){
      fprintf(fp, "    ipaddr = \"%s\"\n", fence_dev->f_ip);
    }
    if(fence_dev->f_login){
      fprintf(fp, "    login = \"%s\"\n", fence_dev->f_login);
    }
    if(fence_dev->f_passwd){
      fprintf(fp, "    passwd = \"%s\"\n", fence_dev->f_passwd);
    }
    fprintf(fp, "  }\n");
  }
  fprintf(fp, "}\n");

  fclose(fp);

  fp = fopen("nodes.ccs", "w+");
  fprintf(fp, "nodes {\n");

  for(lptr = node_list.prev; lptr != &node_list; lptr = lptr->prev){
    node_t *node;
    osi_list_t *lptr2;
    node_fence_t *nf;
    node = osi_list_entry(lptr, node_t, n_list);
    fprintf(fp,
	    "  %s {\n"
	    "    ip_interfaces {\n"
	    "      # Assuming eth0...\n"
	    "      eth0 = \"%s\"\n"
	    "    }\n"
	    "    fence {\n",
	    node->n_name, node->n_ip);
    /* enter node specific fence params */
    for(lptr2 = node->n_flist.prev;
	lptr2 != &node->n_flist;
	lptr2 = lptr2->prev){
      char *fencedev;
      nf = osi_list_entry(lptr2, node_fence_t, nf_list);
      fencedev = fencedev_lookup(nf->nf_name);
      if(fencedev){
	fprintf(fp, "      %s {\n", nf->nf_name);  /* The method name */
	fprintf(fp, "        %s {\n", fencedev); /*dev name*/
	if(strstr(fencedev, "brocade") || strstr(fencedev, "apc") ||
	   strstr(fencedev, "wti") || strstr(fencedev, "vixel") ||
	   strstr(fencedev, "brcd_port")){
	  fprintf(fp, "          port = %s\n", nf->nf_param);
	} else if(strstr(fencedev, "gnbd")){
	  fprintf(fp, "          ipaddr = \"%s\"\n", nf->nf_param);
	} else if(strstr(fencedev, "manual")){
	  fprintf(fp, "          ipaddr = \"%s\"\n", nf->nf_param);
	} else {
	  fprintf(fp, "-- WARNING: unknown_key_for_param = <%s>\n",
		  nf->nf_param);
	}	
	fprintf(fp, "        }\n      }\n");
      } else {
	fprintf(fp, "-- WARNING: Unable to resolve fence device name \"%s\"\n",
		nf->nf_name);
      }
    }      
    fprintf(fp, "    }\n  }\n");
  }

  fprintf(fp, "}\n");
  fclose(fp);
  return 0;
}
	    


int cidev_to_ccs(char *gfs_cf, char *cluster_name){
  FILE *cfg_fp;
  char line[256];
  char param[256];
  node_t *node;
  fence_dev_t *fence_dev;

  cfg_fp = fopen(gfs_cf, "r");

  if(!cfg_fp){
    fprintf(stderr, "Unable to open %s: %s\n", gfs_cf, strerror(errno));
    return -1;
  }


  while(fgets(line, sizeof(line), cfg_fp)){
    if(line[0] == '#'){
      continue;
    } else if(strstr(line, "lmtype")){
      sscanf(line, "lmtype = %s", param);
      /*
      if(!strcmp(param, "dmep")){
	printf("You are using dmep.\n");
      } else {
	printf("You are not using dmep.\n");
      }
      */
      break;
    }
  }

  fseek(cfg_fp, 0, SEEK_SET);

  while(fgets(line, sizeof(line), cfg_fp)){
    if(line[0] == '#'){
      continue;
    } else if(strstr(line, "fence:")){
      fence_dev = malloc(sizeof(fence_dev_t));
      memset(fence_dev, 0, sizeof(fence_dev_t));
      fill_fence(fence_dev, cfg_fp);
      osi_list_add(&fence_dev->f_list, &fence_list);
    }
  }

  fseek(cfg_fp, 0, SEEK_SET);

  while(fgets(line, sizeof(line), cfg_fp)){
    if(line[0] == '#'){
      continue;
    } else if(strstr(line, "node:")){
      node = malloc(sizeof(node_t));
      memset(node, 0, sizeof(node_t));
      osi_list_init(&node->n_flist);
      fill_node(node, cfg_fp);
      osi_list_add(&node->n_list, &node_list);
    }
  }

  fclose(cfg_fp);

  if(generate_files(cluster_name)){
    return -1;
  }

  while(!osi_list_empty(&fence_list)){
    osi_list_del(fence_list.next);
  }
  while(!osi_list_empty(&node_list)){
    node = osi_list_entry(node_list.next, node_t, n_list);
    while(!osi_list_empty(&node->n_flist)){
      osi_list_del(node->n_flist.next);
    }
    osi_list_del(node_list.next);
  }
  /* free lists */
  return 0;
}
