/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#include <linux/pagemap.h>
#include <linux/file.h>
#include <linux/iobuf.h>

#include "bmap.h"
#include "linux_dcache.h"
#include "dio.h"
#include "dir.h"
#include "file.h"
#include "flock.h"
#include "glock.h"
#include "glops.h"
#include "inode.h"
#include "linux_inode.h"
#include "ioctl.h"
#include "log.h"
#include "linux_page.h"
#include "plock.h"
#include "quota.h"
#include "rgrp.h"
#include "trans.h"

#define KIO_MAX_SIZE (KIO_MAX_ATOMIC_IO << 10)

struct filldir_bad_entry {
	char *fbe_name;
	unsigned int fbe_length;
	uint64_t fbe_offset;
	struct gfs_inum fbe_inum;
	unsigned int fbe_type;
};

struct filldir_bad {
	struct gfs_sbd *fdb_sbd;
	int fdb_prefetch;

	struct filldir_bad_entry *fdb_entry;
	unsigned int fdb_entry_num;
	unsigned int fdb_entry_off;

	char *fdb_name;
	unsigned int fdb_name_size;
	unsigned int fdb_name_off;
};

struct filldir_reg {
	struct gfs_sbd *fdr_sbd;
	int fdr_prefetch;

	filldir_t fdr_filldir;
	void *fdr_opaque;
};

typedef ssize_t(*do_rw_t) (struct file * file,
			   char *buf,
			   size_t size, loff_t * offset,
			   unsigned int num_gh, struct gfs_holder * ghs);

struct dio_sizes {
	unsigned int ds_bsize;
	unsigned int ds_bsize_shift;
	unsigned int ds_fsb2b;
	unsigned int ds_fsb2b_shift;
};

/**
 * gfs_llseek - seek to a location in a file
 * @file: the file
 * @offset: the offset
 * @origin: Where to seek from (SEEK_SET, SEEK_CUR, or SEEK_END)
 *
 * SEEK_END requires the glock for the file because it references the
 * file's size.
 *
 * Returns: The new offset, or -EXXX on error
 */

static loff_t
gfs_llseek(struct file *file, loff_t offset, int origin)
{
	struct gfs_inode *ip =
	    (struct gfs_inode *) file->f_dentry->d_inode->u.generic_ip;
	struct gfs_holder i_gh;
	loff_t error;

	if (origin == 2) {
		error =
		    gfs_glock_nq_init(ip->i_gl, LM_ST_SHARED, LM_FLAG_ANY,
				      &i_gh);
		if (!error) {
			error = generic_file_llseek(file, offset, origin);
			gfs_glock_dq_uninit(&i_gh);
		}
	} else
		error = generic_file_llseek(file, offset, origin);

	return error;
}

/**
 * gfs_walk_vma - Walk the vmas associated with a buffer for read or write.
 *    If any of them are gfs, pass the gfs inode down to the read/write
 *    worker function so that locks can be acquired in the correct order.
 * @file: The file to read/write from/to
 * @buf: The buffer to copy to/from
 * @size: The amount of data requested
 * @offset: The current file offset
 * @operation: The read or write worker function
 *
 * Outputs: Offset - updated according to number of bytes written
 *
 * Returns: The number of bytes written, -EXXX on failure
 */

static ssize_t
gfs_walk_vma(struct file *file, char *buf, size_t size,
	     loff_t * offset, do_rw_t operation)
{
	struct vm_area_struct *vma;
	unsigned long start = (unsigned long) buf;
	unsigned long end = start + size;
	struct gfs_holder gh_arr[2], *ghs = gh_arr;
	unsigned int num_gh = 0;
	struct gfs_glock *gl;
	unsigned int state;
	unsigned int x = 0;
	ssize_t count;
	int dumping_core = (current->flags & PF_DUMPCORE) ? TRUE : FALSE;

	/* Check for edge cases before we walk vma list */

	if (end < start) {
		return -EINVAL;
	}
	if (end == start) {
		return 0;
	}

	/*  Handle the case where we're reading in parts of
	   /sbin/init when we're a root FS.  */

	if (!current->mm)
		goto no_mm;

	if (!dumping_core)
		down_read(&current->mm->mmap_sem);

	for (vma = find_vma(current->mm, start); vma; vma = vma->vm_next) {
		if (end <= vma->vm_start)
			break;

		if (vma->vm_file &&
		    vma->vm_file->f_dentry->d_inode->i_sb ==
		    file->f_dentry->d_inode->i_sb && !num_gh++) {
			gl = ((struct gfs_inode *) vma->vm_file->f_dentry->
			      d_inode->u.generic_ip)->i_gl;
			state =
			    ((vma->vm_flags & (VM_MAYWRITE | VM_SHARED)) ==
			     (VM_MAYWRITE | VM_SHARED)) ? LM_ST_EXCLUSIVE :
			    LM_ST_SHARED;
			gfs_init_holder(gl, state, GL_DATA_B, ghs);
		}
	}

	if (num_gh > 1) {
		gfs_uninit_holder(ghs);
		goto handle_multi_lock;
	}

	if (!dumping_core)
		up_read(&current->mm->mmap_sem);

	count = operation(file, buf, size, offset, num_gh, ghs);

	if (num_gh)
		gfs_uninit_holder(ghs);

      out:

	return count;

      handle_multi_lock:
	ghs = gmalloc(sizeof (struct gfs_holder) * (num_gh + 1));

	for (vma = find_vma(current->mm, start); vma; vma = vma->vm_next) {
		if (end <= vma->vm_start)
			break;

		if (vma->vm_file &&
		    vma->vm_file->f_dentry->d_inode->i_sb ==
		    file->f_dentry->d_inode->i_sb) {
			gl = ((struct gfs_inode *) vma->vm_file->f_dentry->
			      d_inode->u.generic_ip)->i_gl;
			state =
			    ((vma->vm_flags & (VM_MAYWRITE | VM_SHARED)) ==
			     (VM_MAYWRITE | VM_SHARED)) ? LM_ST_EXCLUSIVE :
			    LM_ST_SHARED;
			gfs_init_holder(gl, state, GL_DATA_B, &ghs[x]);
			x++;
		}
	}
	GFS_ASSERT_INODE(x == num_gh,
			 (struct gfs_inode *) file->f_dentry->d_inode->u.
			 generic_ip,);

	if (!dumping_core)
		up_read(&current->mm->mmap_sem);

	count = operation(file, buf, size, offset, num_gh, ghs);

	for (x = 0; x < num_gh; x++)
		gfs_uninit_holder(&ghs[x]);

	kfree(ghs);

	goto out;

      no_mm:
	count = operation(file, buf, size, offset, num_gh, ghs);

	goto out;
}

/**
 * find_dio_sizes - figure out what the alignment of this Direct I/O is
 * @sdp: the filesystem
 * @buf: the buffer
 * @offset: the offset in the file
 * @size: the size of the I/O
 * @size: the structure that we return
 *
 * Returns: 0 on success, -EINVAL if the I/O isn't properly aligned
 */

static int
find_dio_sizes(struct gfs_sbd *sdp, char *buf, loff_t offset, size_t size,
	       struct dio_sizes *sizes)
{
	unsigned int fs_mask = sdp->sd_sb.sb_bsize - 1;
	unsigned int bb_mask = GFS_BASIC_BLOCK - 1;
	int error = 0;

	if (!(((unsigned long) buf) & fs_mask) &&
	    !(offset & fs_mask) && !(size & fs_mask)) {
		sizes->ds_bsize = sdp->sd_sb.sb_bsize;
		sizes->ds_bsize_shift = sdp->sd_sb.sb_bsize_shift;
		sizes->ds_fsb2b = 1;
		sizes->ds_fsb2b_shift = 0;
	} else if (GFS_BASIC_BLOCK >= sdp->sd_device_bsize &&
		   !(((unsigned long) buf) & bb_mask) &&
		   !(offset & bb_mask) && !(size & bb_mask)) {
		sizes->ds_bsize = GFS_BASIC_BLOCK;
		sizes->ds_bsize_shift = GFS_BASIC_BLOCK_SHIFT;
		sizes->ds_fsb2b = sdp->sd_fsb2bb;
		sizes->ds_fsb2b_shift = sdp->sd_fsb2bb_shift;
	} else
		error = -EINVAL;

	return error;
}

/**
 * do_direct - do a direct read or non-allocating write
 * @file: the file
 * @buf: the memory that does/will contain the data
 * @size: the size of the I/O
 * @offset: the offset in the file for the I/O
 * @rw: READ or WRITE
 * @sizes: the structure describing the block size of the I/O
 *
 * Function description
 *
 * Returns: what is returned
 */

static ssize_t
do_direct(struct file *file, char *buf, size_t size, loff_t * offset,
	  int rw, struct dio_sizes *sizes)
{
	struct gfs_inode *ip = (struct gfs_inode *) file->f_dentry->d_inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	struct kiobuf *iobuf;
	unsigned int s, s_blocks, x, o;
	uint64_t lblock, dblock;
	uint32_t extlen = 0;
	int not_new = FALSE;
	ssize_t count = 0;
	int error = 0;

	if (gfs_is_stuffed(ip)) {
		GFS_ASSERT(rw == READ,);
		count = gfs_readi(ip, buf, *offset, size, gfs_copy2user);
		if (count > 0)
			*offset += count;
		goto out;
	}

	if (test_and_set_bit(0, &file->f_iobuf_lock)) {
		error = alloc_kiovec(1, &iobuf);
		if (error)
			goto out;
	} else
		iobuf = file->f_iobuf;

	lblock = *offset >> sizes->ds_bsize_shift;

	while (size) {
		s = (size < KIO_MAX_SIZE) ? size : KIO_MAX_SIZE;
		s_blocks = s >> sizes->ds_bsize_shift;

		error = map_user_kiobuf(rw, iobuf, (unsigned long) buf, s);
		if (error)
			goto out_free;

		for (x = 0; x < s_blocks; x++) {
			if (!extlen) {
				error =
				    gfs_block_map(ip,
						  lblock >> sizes->
						  ds_fsb2b_shift, &not_new,
						  &dblock, &extlen);
				if (error) {
					unmap_kiobuf(iobuf);
					goto out_free;
				}

				if (dblock) {
					o = lblock & (sizes->ds_fsb2b - 1);
					dblock =
					    (dblock << sizes->ds_fsb2b_shift) +
					    o;
					extlen =
					    (extlen << sizes->ds_fsb2b_shift) -
					    o;
				} else
					GFS_ASSERT(rw == READ,);
			}

			if (dblock) {
				iobuf->blocks[x] = dblock;
				dblock++;
				extlen--;
			} else
				iobuf->blocks[x] = -1ul;

			lblock++;
		}

		error =
		    brw_kiovec(rw, 1, &iobuf, sdp->sd_vfs->s_dev, iobuf->blocks,
			       sizes->ds_bsize);
		if (error < 0) {
			unmap_kiobuf(iobuf);
			goto out_free;
		}

		if (rw == READ) {
			/*  Handle marking pages dirty  */
		}
		unmap_kiobuf(iobuf);

		*offset += error;
		buf += error;
		size -= error;
		count += error;

		if (error != s) {
			lblock = *offset >> sizes->ds_bsize_shift;
			extlen = 0;
		}
	}

      out_free:
	if (iobuf == file->f_iobuf)
		clear_bit(0, &file->f_iobuf_lock);
	else
		free_kiovec(1, &iobuf);

      out:

	return (count) ? count : error;
}

/**
 * do_read_direct - Read bytes from a file
 * @file: The file to read from
 * @buf: The buffer to copy into
 * @size: The amount of data requested
 * @offset: The current file offset
 * @num_gh: The number of other locks we need to do the read
 * @ghs: the locks we need plus one for our lock
 *
 * Outputs: Offset - updated according to number of bytes read
 *
 * Returns: The number of bytes read, -EXXX on failure
 */

static ssize_t
do_read_direct(struct file *file,
	       char *buf,
	       size_t size, loff_t * offset,
	       unsigned int num_gh, struct gfs_holder *ghs)
{
	struct gfs_inode *ip = (struct gfs_inode *) file->f_dentry->d_inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	struct dio_sizes sizes;
	unsigned int state = LM_ST_DEFERRED;
	int flags = GL_DATA_DR;
	unsigned int x;
	ssize_t count = 0;
	int error;

	for (x = 0; x < num_gh; x++)
		if (ghs[x].gh_gl == ip->i_gl) {
			state = LM_ST_SHARED;
			flags |= GL_LOCAL_EXCL;
			break;
		}

	gfs_init_holder(ip->i_gl, state, flags, &ghs[num_gh++]);

	error = gfs_glock_nq_m(num_gh, ghs);
	if (error)
		goto out;

	if (gfs_is_jdata(ip)) {
		error = -EINVAL;
		goto out_gunlock;
	}

	if (!(file->f_flags & O_LARGEFILE)) {
		if (*offset >= 0x7FFFFFFFull) {
			error = -EFBIG;
			goto out_gunlock;
		}
		if (*offset + size > 0x7FFFFFFFull)
			size = 0x7FFFFFFFull - *offset;
	}

	if (*offset >= ip->i_di.di_size)
		goto out_gunlock;

	if ((*offset + size) > ip->i_di.di_size) {
		if (size & (GFS_BASIC_BLOCK - 1)) {
			error = -EINVAL;
			goto out_gunlock;
		}
		size = ip->i_di.di_size - *offset;
		size &= ~(size_t) (GFS_BASIC_BLOCK - 1);
		if (!size)
			goto out_gunlock;
	}

	error = find_dio_sizes(sdp, buf, *offset, size, &sizes);
	if (error)
		goto out_gunlock;

	count = do_direct(file, buf, size, offset, READ, &sizes);

      out_gunlock:
	gfs_glock_dq_m(num_gh, ghs);

      out:
	gfs_uninit_holder(&ghs[num_gh - 1]);

	return (count) ? count : error;
}

/**
 * do_read - Read bytes from a file
 * @file: The file to read from
 * @buf: The buffer to copy into
 * @size: The amount of data requested
 * @offset: The current file offset
 * @num_gh: The number of other locks we need to do the read
 * @ghs: the locks we need plus one for our lock
 *
 * Outputs: Offset - updated according to number of bytes read
 *
 * Returns: The number of bytes read, -EXXX on failure
 */

static ssize_t
do_read(struct file *file,
	char *buf,
	size_t size, loff_t * offset, unsigned int num_gh,
	struct gfs_holder *ghs)
{
	struct gfs_inode *ip = (struct gfs_inode *) file->f_dentry->d_inode->u.generic_ip;
	ssize_t count = 0;
	int error;

	gfs_init_holder(ip->i_gl, LM_ST_SHARED, GL_ATIME | GL_DATA_B,
			&ghs[num_gh++]);

	error = gfs_glock_nq_m_atime(num_gh, ghs);
	if (error)
		goto out;

	if (gfs_is_jdata(ip) ||
	    (gfs_is_stuffed(ip) && !test_bit(GIF_PAGED, &ip->i_flags))) {
		if (!(file->f_flags & O_LARGEFILE)) {
			if (*offset >= 0x7FFFFFFFull) {
				error = -EFBIG;
				goto out_gunlock;
			}
			if (*offset + size > 0x7FFFFFFFull)
				size = 0x7FFFFFFFull - *offset;
		}

		count = gfs_readi(ip, buf, *offset, size, gfs_copy2user);

		if (count > 0)
			*offset += count;
	} else
		count = generic_file_read(file, buf, size, offset);

      out_gunlock:
	gfs_glock_dq_m(num_gh, ghs);

      out:
	gfs_uninit_holder(&ghs[num_gh - 1]);

	return (count) ? count : error;
}

/**
 * gfs_read - Read bytes from a file
 * @file: The file to read from
 * @buf: The buffer to copy into
 * @size: The amount of data requested
 * @offset: The current file offset
 *
 * Outputs: Offset - updated according to number of bytes read
 *
 * Returns: The number of bytes read, -EXXX on failure
 */

static ssize_t
gfs_read(struct file *file, char *buf, size_t size, loff_t * offset)
{
	ssize_t count;

	if (!size)
		return 0;
	if (*offset < 0)
		return -EINVAL;
	if (!access_ok(VERIFY_WRITE, buf, size))
		return -EFAULT;

	if (file->f_flags & O_DIRECT)
		count = gfs_walk_vma(file, buf, size, offset, do_read_direct);
	else
		count = gfs_walk_vma(file, buf, size, offset, do_read);

	return count;
}

/**
 * grope_mapping - feel up a mapping that needs to be written
 * @buf: the start of the memory to be written
 * @size: the size of the memory to be written
 *
 * We do this after acquiring the locks on the mapping,
 * but before starting the write transaction.  We need to make
 * sure that we don't cause recursive transactions if blocks
 * need to be allocated to the file backing the mapping.
 *
 * Returns:  0 on success, -EXXX on failure
 */

static int
grope_mapping(char *buf, size_t size)
{
	unsigned long start = (unsigned long) buf;
	unsigned long stop = start + size;
	char c;
	int error = 0;

	while (start < stop) {
		error = (copy_from_user(&c, (char *) start, 1)) ? -EFAULT : 0;
		if (error)
			break;

		start += PAGE_CACHE_SIZE;
		start &= PAGE_CACHE_MASK;
	}

	return error;
}

/**
 * do_direct_alloc - do a direct write that allocates blocks
 * @file: the file
 * @buf: the memory that does/will contain the data
 * @size: the size of the I/O
 * @offset: the offset in the file for the I/O
 * @sizes: the structure describing the block size of the I/O
 *
 * Function description
 *
 * Returns: what is returned
 */

static ssize_t
do_direct_alloc(struct file *file, char *buf, size_t size, loff_t * offset,
		struct dio_sizes *sizes)
{
	struct gfs_inode *ip = (struct gfs_inode *) file->f_dentry->d_inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	struct kiobuf *iobuf;
	struct buffer_head *bh;
	int write_head = FALSE, write_tail = FALSE;
	unsigned int s, s_blocks, x, o;
	uint64_t lblock, dblock, lblock_fs, lstop_fs;
	uint32_t extlen = 0;
	int new, first = TRUE;
	ssize_t count = 0;
	int error = 0;

	if (sizes->ds_bsize != sdp->sd_sb.sb_bsize) {
		if (*offset & (sdp->sd_sb.sb_bsize - 1))
			write_head = TRUE;
		if ((*offset + size) & (sdp->sd_sb.sb_bsize - 1))
			write_tail = TRUE;
	}

	if (gfs_is_stuffed(ip)) {
		error = gfs_unstuff_dinode(ip, gfs_unstuffer_sync, NULL);
		if (error)
			goto out;
	}

	if (test_and_set_bit(0, &file->f_iobuf_lock)) {
		error = alloc_kiovec(1, &iobuf);
		if (error)
			goto out;
	} else
		iobuf = file->f_iobuf;

	lblock = *offset >> sizes->ds_bsize_shift;
	lstop_fs = (*offset + size - 1) >> sdp->sd_sb.sb_bsize_shift;

	while (size) {
		GFS_ASSERT_INODE(size <= KIO_MAX_SIZE, ip,);
		s = size;
		s_blocks = s >> sizes->ds_bsize_shift;

		error = map_user_kiobuf(WRITE, iobuf, (unsigned long) buf, s);
		if (error)
			goto out_free;

		for (x = 0; x < s_blocks; x++) {
			if (!extlen) {
				new = TRUE;
				lblock_fs = lblock >> sizes->ds_fsb2b_shift;

				error =
				    gfs_block_map(ip, lblock_fs, &new, &dblock,
						  &extlen);
				if (error) {
					unmap_kiobuf(iobuf);
					goto out_free;
				}

				GFS_ASSERT_INODE(dblock, ip,);

				if (new && ((first && write_head)
					    || (lblock_fs == lstop_fs
						&& write_tail))) {
					bh = gfs_dgetblk(sdp, dblock, ip->i_gl);
					gfs_prep_new_buffer(bh);
					gfs_buffer_clear(bh);
					error =
					    gfs_dwrite(sdp, bh,
						       DIO_DIRTY | DIO_START |
						       DIO_WAIT);
					gfs_drelse(bh);
					if (error) {
						unmap_kiobuf(iobuf);
						goto out_free;
					}
				}

				o = lblock & (sizes->ds_fsb2b - 1);
				dblock = (dblock << sizes->ds_fsb2b_shift) + o;
				extlen = (extlen << sizes->ds_fsb2b_shift) - o;
			}

			iobuf->blocks[x] = dblock;
			dblock++;
			extlen--;
			lblock++;
			first = FALSE;
		}

		error =
		    brw_kiovec(WRITE, 1, &iobuf, sdp->sd_vfs->s_dev,
			       iobuf->blocks, sizes->ds_bsize);
		if (error < 0) {
			unmap_kiobuf(iobuf);
			goto out_free;
		}

		unmap_kiobuf(iobuf);

		*offset += error;
		buf += error;
		size -= error;
		count += error;

		if (error != s) {
			lblock = *offset >> sizes->ds_bsize_shift;
			extlen = 0;
		}
	}

      out_free:
	if (iobuf == file->f_iobuf)
		clear_bit(0, &file->f_iobuf_lock);
	else
		free_kiovec(1, &iobuf);

      out:

	return (count) ? count : error;
}

/**
 * do_write_direct_alloc - do a Direct write that requires allocation
 * @buf: the memory that does/will contain the data
 * @size: the size of the I/O
 * @offset: the offset in the file for the I/O
 * @sizes: the structure describing the block size of the I/O
 *
 * Returns: the number of bytes written or -EXXX on failure
 */

static ssize_t
do_write_direct_alloc(struct file *file,
		      char *buf,
		      size_t size, loff_t * offset, struct dio_sizes *sizes)
{
	struct gfs_inode *ip = (struct gfs_inode *) file->f_dentry->d_inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_alloc *al = NULL;
	struct buffer_head *dibh;
	unsigned int data_blocks, ind_blocks;
	ssize_t count;
	int error;

	gfs_write_calc_reserv(ip, size, &data_blocks, &ind_blocks);

	al = gfs_get_alloc(ip);

	error =
	    gfs_get_quota_locks(ip, NO_QUOTA_CHANGE, NO_QUOTA_CHANGE,
				gfs_get_user_cred());
	if (error)
		goto fail;

	al->al_metablks = ind_blocks;
	al->al_datablks = data_blocks;
	al->al_trans = TRANS_WRITE;

	error = gfs_inplace_reserv(ip);
	if (error)
		goto fail_gunlock_q;

	error = gfs_check_quota(ip, ip->i_di.di_uid, ip->i_di.di_gid);
	if (error)
		goto fail_ipres;

	/*  Trans may require:
	   All blocks for a RG bitmap, whatever indirect blocks we
	   need, a modified dinode, and a quota change.  */

	error =
	    gfs_trans_begin(sdp, TRANS_WRITE,
			    1 + al->al_rgd->rd_ri.ri_length + ind_blocks, 1);
	if (error)
		goto fail_ipres;

	count = do_direct_alloc(file, buf, size, offset, sizes);
	if (count < 0) {
		error = count;
		goto fail_end_trans;
	}

	error = gfs_get_inode_buffer(ip, &dibh);
	if (error)
		goto fail_end_trans;

	if (ip->i_di.di_size < *offset)
		ip->i_di.di_size = *offset;
	ip->i_di.di_mtime = ip->i_di.di_ctime = CURRENT_TIME;
	if ((ip->i_di.di_mode & (S_ISUID | S_ISGID)) && !capable(CAP_FSETID))
		ip->i_di.di_mode &=
		    (ip->i_di.
		     di_mode & S_IXGRP) ? (~(S_ISUID | S_ISGID)) : (~S_ISUID);

	gfs_trans_add_bh(sdp, ip->i_gl, dibh);
	gfs_dinode_out(&ip->i_di, (dibh)->b_data);
	gfs_drelse(dibh);

	gfs_trans_end(sdp);

	/*  Don't check al->al_metaalloced and friends.  */
	gfs_inplace_relse(ip);
	gfs_put_quota_locks(ip);
	gfs_put_alloc(ip);

	return count;

      fail_end_trans:
	gfs_trans_end(sdp);

      fail_ipres:
	gfs_inplace_relse(ip);

      fail_gunlock_q:
	gfs_put_quota_locks(ip);

      fail:
	gfs_put_alloc(ip);

	return error;
}

/**
 * do_write_direct - Write bytes to a file
 * @file: The file to write to
 * @buf: The buffer to copy from
 * @size: The amount of data requested
 * @offset: The current file offset
 * @num_gh: The number of other locks we need to do the read
 * @gh: the locks we need plus one for our lock
 *
 * Outputs: Offset - updated according to number of bytes written
 *
 * Returns: The number of bytes written, -EXXX on failure
 */

static ssize_t
do_write_direct(struct file *file,
		char *buf,
		size_t size, loff_t * offset,
		unsigned int num_gh, struct gfs_holder *ghs)
{
	struct gfs_inode *ip = (struct gfs_inode *) file->f_dentry->d_inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_file *fp = gfs_get_file(file);
	struct dio_sizes sizes;
	unsigned int state = LM_ST_DEFERRED;
	struct gfs_holder t_gh;
	int alloc_required;
	unsigned int x;
	size_t s;
	ssize_t count = 0;
	int error;

	GFS_ASSERT(fp,);

	if (test_bit(GFF_DID_DIRECT_ALLOC, &fp->f_flags))
		state = LM_ST_EXCLUSIVE;
	else
		for (x = 0; x < num_gh; x++)
			if (ghs[x].gh_gl == ip->i_gl) {
				state = LM_ST_EXCLUSIVE;
				break;
			}

      restart:
	gfs_init_holder(ip->i_gl, state, GL_DATA_DW, &ghs[num_gh++]);

	error = gfs_glock_nq_m(num_gh, ghs);
	if (error)
		goto out;

	if (gfs_is_jdata(ip)) {
		error = -EINVAL;
		goto out_gunlock;
	}

	if (num_gh > 1) {
		error = grope_mapping(buf, size);
		if (error)
			goto out_gunlock;
	}

	if (file->f_flags & O_APPEND)
		*offset = ip->i_di.di_size;

	if (!(file->f_flags & O_LARGEFILE)) {
		if (*offset >= 0x7FFFFFFFull) {
			error = -EFBIG;
			goto out_gunlock;
		}
		if (*offset + size > 0x7FFFFFFFull)
			size = 0x7FFFFFFFull - *offset;
	}

	error = find_dio_sizes(sdp, buf, *offset, size, &sizes);
	if (error)
		goto out_gunlock;

	if (gfs_is_stuffed(ip) ||
	    *offset + size > ip->i_di.di_size ||
	    ((ip->i_di.di_mode & (S_ISUID | S_ISGID)) && !capable(CAP_FSETID)))
		alloc_required = TRUE;
	else {
		error =
		    gfs_write_alloc_required(ip, *offset, size,
					     &alloc_required);
		if (error)
			goto out_gunlock;
	}

	if (alloc_required && state != LM_ST_EXCLUSIVE) {
		gfs_glock_dq_m(num_gh, ghs);
		num_gh--;
		gfs_uninit_holder(&ghs[num_gh]);
		state = LM_ST_EXCLUSIVE;
		goto restart;
	}

	if (alloc_required) {
		set_bit(GFF_DID_DIRECT_ALLOC, &fp->f_flags);

		while (size) {
			s = sdp->sd_tune.gt_max_atomic_write;
			if (s > size)
				s = size;
			if (s > KIO_MAX_SIZE)
				s = KIO_MAX_SIZE;
			s &= ~(size_t) (sizes.ds_bsize - 1);
			if (s < sizes.ds_bsize)
				s = sizes.ds_bsize;

			error =
			    do_write_direct_alloc(file, buf, s, offset, &sizes);
			if (error < 0)
				goto out_gunlock;

			buf += error;
			size -= error;
			count += error;
		}

		if (file->f_flags & O_SYNC)
			gfs_log_flush_glock(ip->i_gl);
	} else {
		clear_bit(GFF_DID_DIRECT_ALLOC, &fp->f_flags);

		error =
		    gfs_glock_nq_init(sdp->sd_trans_gl, LM_ST_SHARED, 0, &t_gh);
		if (error)
			goto out_gunlock;

		count = do_direct(file, buf, size, offset, WRITE, &sizes);

		gfs_glock_dq_uninit(&t_gh);
	}

      out_gunlock:
	gfs_glock_dq_m(num_gh, ghs);

      out:
	gfs_uninit_holder(&ghs[num_gh - 1]);

	return (count) ? count : error;
}

/**
 * do_do_write - Write bytes to a file
 * @file: The file to write to
 * @buf: The buffer to copy from
 * @size: The amount of data requested
 * @offset: The current file offset
 * @num_gh: The number of locks that were acquired in do_write()
 * @ghs: The locks that were acquired in do_write()
 *
 * Outputs: Offset - updated according to number of bytes written
 *
 * Returns: The number of bytes written, -EXXX on failure
 */

static ssize_t
do_do_write(struct file *file,
	    char *buf,
	    size_t size, loff_t * offset,
	    unsigned int num_gh, struct gfs_holder *ghs)
{
	struct inode *inode = file->f_dentry->d_inode;
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_alloc *al = NULL;
	struct buffer_head *dibh;
	unsigned int data_blocks, ind_blocks;
	int alloc_required, journaled;
	ssize_t count;
	int error;

	journaled = gfs_is_jdata(ip);

	gfs_write_calc_reserv(ip, size, &data_blocks, &ind_blocks);

	error = gfs_write_alloc_required(ip, *offset, size, &alloc_required);
	if (error)
		goto fail;

	if (alloc_required) {
		al = gfs_get_alloc(ip);

		error =
		    gfs_get_quota_locks(ip, NO_QUOTA_CHANGE, NO_QUOTA_CHANGE,
					gfs_get_user_cred());
		if (error)
			goto fail_alloc;

		if (journaled)
			al->al_metablks = ind_blocks + data_blocks;
		else {
			al->al_metablks = ind_blocks;
			al->al_datablks = data_blocks;
		}
		al->al_trans = TRANS_WRITE;

		error = gfs_inplace_reserv(ip);
		if (error)
			goto fail_gunlock_q;

		error = gfs_check_quota(ip, ip->i_di.di_uid, ip->i_di.di_gid);
		if (error)
			goto fail_ipres;

		/*  Trans may require:
		   All blocks for a RG bitmap, whatever indirect blocks we
		   need, a modified dinode, and a quota change.  */

		error =
		    gfs_trans_begin(sdp, TRANS_WRITE,
				    1 + al->al_rgd->rd_ri.ri_length +
				    ind_blocks +
				    ((journaled) ? data_blocks : 0), 1);
		if (error)
			goto fail_ipres;
	} else {
		/*  Trans may require:
		   A modified dinode.  */

		error =
		    gfs_trans_begin(sdp, TRANS_WRITE,
				    1 + ((journaled) ? data_blocks : 0), 0);
		if (error)
			goto fail_ipres;
	}

	if ((ip->i_di.di_mode & (S_ISUID | S_ISGID)) && !capable(CAP_FSETID)) {
		error = gfs_get_inode_buffer(ip, &dibh);
		if (error)
			goto fail_end_trans;

		ip->i_di.di_mode &=
		    (ip->i_di.
		     di_mode & S_IXGRP) ? (~(S_ISUID | S_ISGID)) : (~S_ISUID);

		gfs_trans_add_bh(sdp, ip->i_gl, dibh);
		gfs_dinode_out(&ip->i_di, (dibh)->b_data);
		gfs_drelse(dibh);
	}

	if (journaled ||
	    (gfs_is_stuffed(ip) &&
	     *offset + size <= sdp->sd_sb.sb_bsize - sizeof (struct gfs_dinode)
	     && !test_bit(GIF_PAGED, &ip->i_flags))) {
		count = gfs_writei(ip, buf, *offset, size, gfs_copy_from_user);
		if (count < 0) {
			error = count;
			goto fail_end_trans;
		}

		gfs_trans_end(sdp);

		if ((file->f_flags & O_SYNC) ||
		    test_bit(SDF_SYNC_MOUNT, &sdp->sd_flags)) {
			if (journaled)
				gfs_log_flush_glock(ip->i_gl);
			else
				ghs[num_gh - 1].gh_flags |= GL_SYNC;
		}

		*offset += count;
	} else {
		count = do_generic_file_write(file, buf, size, offset);
		if (count < 0) {
			error = count;
			goto fail_end_trans;
		}

		error = gfs_get_inode_buffer(ip, &dibh);
		if (error)
			goto fail_end_trans;

		if (ip->i_di.di_size < inode->i_size)
			ip->i_di.di_size = inode->i_size;
		ip->i_di.di_mtime = ip->i_di.di_ctime = CURRENT_TIME;

		gfs_trans_add_bh(sdp, ip->i_gl, dibh);
		gfs_dinode_out(&ip->i_di, (dibh)->b_data);
		gfs_drelse(dibh);

		gfs_trans_end(sdp);

		/*  Only need to sync metadata here.
		   gfs_trans_end() syncs metadata on SDF_SYNC_MOUNT  */

		if (file->f_flags & O_SYNC)
			gfs_log_flush_glock(ip->i_gl);
	}

	if (alloc_required) {
		GFS_ASSERT_INODE(count != size || al->al_metaalloced
				 || al->al_dataalloced, ip,);
		gfs_inplace_relse(ip);
		gfs_put_quota_locks(ip);
		gfs_put_alloc(ip);
	}

	return count;

      fail_end_trans:
	gfs_trans_end(sdp);

      fail_ipres:
	if (alloc_required)
		gfs_inplace_relse(ip);

      fail_gunlock_q:
	if (alloc_required)
		gfs_put_quota_locks(ip);

      fail_alloc:
	if (alloc_required)
		gfs_put_alloc(ip);

      fail:
	return error;
}

/**
 * do_write - Write bytes to a file
 * @file: The file to write to
 * @buf: The buffer to copy from
 * @size: The amount of data requested
 * @offset: The current file offset
 * @num_gh: The number of other locks we need to do the read
 * @gh: the locks we need plus one for our lock
 *
 * Outputs: Offset - updated according to number of bytes written
 *
 * Returns: The number of bytes written, -EXXX on failure
 */

static ssize_t
do_write(struct file *file,
	 char *buf,
	 size_t size, loff_t * offset, unsigned int num_gh,
	 struct gfs_holder *ghs)
{
	struct gfs_inode *ip = (struct gfs_inode *) file->f_dentry->d_inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	size_t s;
	ssize_t count = 0;
	int error;

	gfs_init_holder(ip->i_gl, LM_ST_EXCLUSIVE, GL_DATA_B, &ghs[num_gh++]);

	error = gfs_glock_nq_m(num_gh, ghs);
	if (error)
		goto out;

	if (num_gh > 1) {
		error = grope_mapping(buf, size);
		if (error)
			goto out_gunlock;
	}

	if (file->f_flags & O_APPEND)
		*offset = ip->i_di.di_size;

	if (!(file->f_flags & O_LARGEFILE)) {
		if (*offset >= 0x7FFFFFFFull) {
			error = -EFBIG;
			goto out_gunlock;
		}
		if (*offset + size > 0x7FFFFFFFull)
			size = 0x7FFFFFFFull - *offset;
	}

	while (size) {
		s = sdp->sd_tune.gt_max_atomic_write;
		if (s > size)
			s = size;

		error = do_do_write(file, buf, s, offset, num_gh, ghs);
		if (error < 0)
			goto out_gunlock;

		buf += error;
		size -= error;
		count += error;
	}

      out_gunlock:
	gfs_glock_dq_m(num_gh, ghs);

      out:
	gfs_uninit_holder(&ghs[num_gh - 1]);

	return (count) ? count : error;
}

/**
 * gfs_write - Write bytes to a file
 * @file: The file to write to
 * @buf: The buffer to copy from
 * @size: The amount of data requested
 * @offset: The current file offset
 *
 * Outputs: Offset - updated according to number of bytes written
 *
 * Returns: The number of bytes written, -EXXX on failure
 */

static ssize_t
gfs_write(struct file *file, const char *buf, size_t size, loff_t * offset)
{
	struct inode *inode = file->f_dentry->d_inode;
	ssize_t count;

	if (!size)
		return 0;
	if (*offset < 0)
		return -EINVAL;
	if (!access_ok(VERIFY_READ, buf, size))
		return -EFAULT;

	if (file->f_flags & O_DIRECT) {
		down(&inode->i_sem);
		count =
		    gfs_walk_vma(file, (char *) buf, size, offset,
				 do_write_direct);
		up(&inode->i_sem);
	} else {
		down(&inode->i_sem);
		count =
		    gfs_walk_vma(file, (char *) buf, size, offset, do_write);
		up(&inode->i_sem);
	}

	return count;
}

/**
 * filldir_reg_func - Report a directory entry to the caller of gfs_dir_read()
 * @opaque: opaque data used by the function
 * @name: the name of the directory entry
 * @length: the length of the name
 * @offset: the entry's offset in the directory
 * @inum: the inode number the entry points to
 * @type: the type of inode the entry points to
 *
 * Returns: 0 on success, 1 if buffer full
 */

static int
filldir_reg_func(void *opaque,
		 const char *name, unsigned int length,
		 uint64_t offset, struct gfs_inum *inum, unsigned int type)
{
	struct filldir_reg *fdr = (struct filldir_reg *) opaque;
	struct gfs_sbd *sdp = fdr->fdr_sbd;
	unsigned int vfs_type;
	int error;

	switch (type) {
	case GFS_FILE_NON:
		vfs_type = DT_UNKNOWN;
		break;
	case GFS_FILE_REG:
		vfs_type = DT_REG;
		break;
	case GFS_FILE_DIR:
		vfs_type = DT_DIR;
		break;
	case GFS_FILE_LNK:
		vfs_type = DT_LNK;
		break;
	case GFS_FILE_BLK:
		vfs_type = DT_BLK;
		break;
	case GFS_FILE_CHR:
		vfs_type = DT_CHR;
		break;
	case GFS_FILE_FIFO:
		vfs_type = DT_FIFO;
		break;
	case GFS_FILE_SOCK:
		vfs_type = DT_SOCK;
		break;
	default:
		GFS_ASSERT_SBD(FALSE, sdp, printk("type = %u\n", type);
		    );
	}

	error =
	    fdr->fdr_filldir(fdr->fdr_opaque, name, length, offset,
			     inum->no_formal_ino, vfs_type);
	if (error)
		return 1;

	if (fdr->fdr_prefetch && !(length == 1 && *name == '.')) {
		gfs_glock_prefetch_num(sdp, inum->no_formal_ino,
				       &gfs_inode_glops, LM_ST_SHARED,
				       LM_FLAG_TRY | LM_FLAG_ANY);
		gfs_glock_prefetch_num(sdp, inum->no_addr, &gfs_iopen_glops,
				       LM_ST_SHARED, LM_FLAG_TRY);
	}

	return 0;
}

/**
 * readdir_reg - Read directory entries from a directory
 * @file: The directory to read from
 * @dirent: Buffer for dirents
 * @filldir: Function used to do the copying
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
readdir_reg(struct file *file, void *dirent, filldir_t filldir)
{
	struct gfs_inode *dip = (struct gfs_inode *) file->f_dentry->d_inode->u.generic_ip;
	struct filldir_reg fdr;
	struct gfs_holder d_gh;
	uint64_t offset = file->f_pos;
	int error;

	fdr.fdr_sbd = dip->i_sbd;
	fdr.fdr_prefetch = GFS_ASYNC_LM(dip->i_sbd);
	fdr.fdr_filldir = filldir;
	fdr.fdr_opaque = dirent;

	gfs_init_holder(dip->i_gl, LM_ST_SHARED, GL_ATIME, &d_gh);
	error = gfs_glock_nq_atime(&d_gh);
	if (error) {
		gfs_uninit_holder(&d_gh);
		return error;
	}

	error = gfs_dir_read(dip, &offset, &fdr, filldir_reg_func);

	gfs_glock_dq_uninit(&d_gh);

	file->f_pos = offset;

	return error;
}

/**
 * filldir_bad_func - Report a directory entry to the caller of gfs_dir_read()
 * @opaque: opaque data used by the function
 * @name: the name of the directory entry
 * @length: the length of the name
 * @offset: the entry's offset in the directory
 * @inum: the inode number the entry points to
 * @type: the type of inode the entry points to
 *
 * Returns: 0 on success, 1 if buffer full
 */

static int
filldir_bad_func(void *opaque,
		 const char *name, unsigned int length,
		 uint64_t offset, struct gfs_inum *inum, unsigned int type)
{
	struct filldir_bad *fdb = (struct filldir_bad *) opaque;
	struct gfs_sbd *sdp = fdb->fdb_sbd;
	struct filldir_bad_entry *fbe;

	if (fdb->fdb_entry_off == fdb->fdb_entry_num ||
	    fdb->fdb_name_off + length > fdb->fdb_name_size)
		return 1;

	fbe = &fdb->fdb_entry[fdb->fdb_entry_off];
	fbe->fbe_name = fdb->fdb_name + fdb->fdb_name_off;
	memcpy(fbe->fbe_name, name, length);
	fbe->fbe_length = length;
	fbe->fbe_offset = offset;
	fbe->fbe_inum = *inum;
	fbe->fbe_type = type;

	fdb->fdb_entry_off++;
	fdb->fdb_name_off += length;

	if (fdb->fdb_prefetch && !(length == 1 && *name == '.')) {
		gfs_glock_prefetch_num(sdp, inum->no_formal_ino,
				       &gfs_inode_glops, LM_ST_SHARED,
				       LM_FLAG_TRY | LM_FLAG_ANY);
		gfs_glock_prefetch_num(sdp, inum->no_addr, &gfs_iopen_glops,
				       LM_ST_SHARED, LM_FLAG_TRY);
	}

	return 0;
}

/**
 * readdir_bad - Read directory entries from a directory
 * @file: The directory to read from
 * @dirent: Buffer for dirents
 * @filldir: Function used to do the copying
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
readdir_bad(struct file *file, void *dirent, filldir_t filldir)
{
	struct gfs_inode *dip = (struct gfs_inode *) file->f_dentry->d_inode->u.generic_ip;
	struct gfs_sbd *sdp = dip->i_sbd;
	struct filldir_reg fdr;
	unsigned int entries, size;
	struct filldir_bad *fdb;
	struct gfs_holder d_gh;
	uint64_t offset = file->f_pos;
	unsigned int x;
	struct filldir_bad_entry *fbe;
	int error;

	entries = sdp->sd_tune.gt_entries_per_readdir;
	size =
	    sizeof (struct filldir_bad) +
	    entries * (sizeof (struct filldir_bad_entry) + GFS_FAST_NAME_SIZE);

	fdb = gmalloc(size);
	memset(fdb, 0, size);

	fdb->fdb_sbd = sdp;
	fdb->fdb_prefetch = GFS_ASYNC_LM(sdp);
	fdb->fdb_entry = (struct filldir_bad_entry *) (fdb + 1);
	fdb->fdb_entry_num = entries;
	fdb->fdb_name =
	    ((char *) fdb) + sizeof (struct filldir_bad) +
	    entries * sizeof (struct filldir_bad_entry);
	fdb->fdb_name_size = entries * GFS_FAST_NAME_SIZE;

	gfs_init_holder(dip->i_gl, LM_ST_SHARED, GL_ATIME, &d_gh);
	error = gfs_glock_nq_atime(&d_gh);
	if (error) {
		gfs_uninit_holder(&d_gh);
		goto out;
	}

	error = gfs_dir_read(dip, &offset, fdb, filldir_bad_func);

	gfs_glock_dq_uninit(&d_gh);

	fdr.fdr_sbd = sdp;
	fdr.fdr_prefetch = FALSE;
	fdr.fdr_filldir = filldir;
	fdr.fdr_opaque = dirent;

	for (x = 0; x < fdb->fdb_entry_off; x++) {
		fbe = &fdb->fdb_entry[x];

		error =
		    filldir_reg_func(&fdr, fbe->fbe_name, fbe->fbe_length,
				     fbe->fbe_offset, &fbe->fbe_inum,
				     fbe->fbe_type);
		if (error) {
			file->f_pos = fbe->fbe_offset;
			error = 0;
			goto out;
		}
	}

	file->f_pos = offset;

      out:
	kfree(fdb);

	return error;
}

/**
 * gfs_readdir - Read directory entries from a directory
 * @file: The directory to read from
 * @dirent: Buffer for dirents
 * @filldir: Function used to do the copying
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
gfs_readdir(struct file *file, void *dirent, filldir_t filldir)
{
	int error;

	if (strcmp(current->comm, "nfsd") != 0)
		error = readdir_reg(file, dirent, filldir);
	else
		error = readdir_bad(file, dirent, filldir);

	return error;
}

/**
 * gfs_ioctl - do an ioctl on a file
 * @inode: the inode
 * @file: the file pointer
 * @cmd: the ioctl command
 * @arg: the argument
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
gfs_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	  unsigned long arg)
{
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;
	int error = 0;

	GFS_ASSERT(ip,);

	error = gfs_ioctli(ip, cmd, (void *) arg, gfs_get_user_cred());

	return error;
}

/**
 * gfs_open - open a file
 * @inode: the inode to open
 * @file: the struct file for this opening
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
gfs_open(struct inode *inode, struct file *file)
{
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;
	struct gfs_holder i_gh;
	struct gfs_file *fp;
	int error;

	GFS_ASSERT_INODE(file->f_dentry->d_inode == inode, ip,);

	fp = gmalloc(sizeof (struct gfs_file));
	memset(fp, 0, sizeof (struct gfs_file));

	init_MUTEX(&fp->f_fl_lock);

	fp->f_inode = ip;
	fp->f_vfile = file;

	GFS_ASSERT_INODE(!gfs_get_file(file), ip,);
	gfs_set_file(file, fp);

	if (ip->i_di.di_type == GFS_FILE_REG) {
		error =
		    gfs_glock_nq_init(ip->i_gl, LM_ST_SHARED, LM_FLAG_ANY,
				      &i_gh);
		if (error)
			goto fail;

		if (!(file->f_flags & O_LARGEFILE)
		    && ip->i_di.di_size > 0x7FFFFFFFull) {
			error = -EFBIG;
			goto fail_gunlock;
		}

		/*  If this is an exclusive create, make sure our gfs_create()
		   says we created the file.  The O_EXCL flag isn't passed
		   to gfs_create(), so we have to check it here.  */

		if (file->f_flags & O_CREAT) {
			if (ip->i_arch.i_creat_task == current
			    && ip->i_arch.i_creat_pid == current->pid) {
				ip->i_arch.i_creat_task = NULL;
				ip->i_arch.i_creat_pid = 0;
			} else if (file->f_flags & O_EXCL) {
				error = -EEXIST;
				goto fail_gunlock;
			}
		}

		/*  Listen to the Direct I/O flag  */

		if (!(file->f_flags & O_DIRECT)
		    && (ip->i_di.di_flags & GFS_DIF_DIRECTIO)) {
			error = alloc_kiovec(1, &file->f_iobuf);
			if (error)
				goto fail_gunlock;
			file->f_flags |= O_DIRECT;
		}

		/*  Don't let the user open O_DIRECT on a jdata file  */

		if ((file->f_flags & O_DIRECT) && gfs_is_jdata(ip)) {
			error = -EINVAL;
			goto fail_gunlock;
		}

		gfs_glock_dq_uninit(&i_gh);
	}

	return 0;

      fail_gunlock:
	gfs_glock_dq_uninit(&i_gh);

      fail:
	gfs_set_file(file, NULL);
	kfree(fp);

	return error;
}

/**
 * gfs_close - called to close a struct file
 * @inode: the inode the struct file belongs to
 * @file: the struct file being closed
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
gfs_close(struct inode *inode, struct file *file)
{
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;
	struct gfs_file *fp;

	GFS_ASSERT_INODE(file->f_dentry->d_inode == inode, ip,);

	fp = gfs_get_file(file);
	gfs_set_file(file, NULL);

	GFS_ASSERT(fp,);

	kfree(fp);

	return 0;
}

/**
 * gfs_sync_file - sync the dirty data for a file (across the cluster)
 * @file: the file that points to the dentry (Huh?)
 * @dentry: the dentry that points to the inode to sync
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
gfs_sync_file(struct file *file, struct dentry *dentry, int datasync)
{
	struct gfs_inode *ip =
	    (struct gfs_inode *) dentry->d_inode->u.generic_ip;
	struct gfs_holder i_gh;
	int error;

	error = gfs_glock_nq_init(ip->i_gl, LM_ST_SHARED, LM_FLAG_ANY, &i_gh);
	if (error)
		goto out;

	if (gfs_is_jdata(ip))
		gfs_log_flush_glock(ip->i_gl);
	else
		i_gh.gh_flags |= GL_SYNC;

	gfs_glock_dq_uninit(&i_gh);

      out:

	return error;
}

/**
 * gfs_lock_file - acquire/release a flock or posix lock on a file
 * @file: the file pointer
 * @cmd: either modify or retrieve lock state, possibly wait
 * @fl: type and range of lock
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
gfs_lock_file(struct file *file, int cmd, struct file_lock *fl)
{
	struct gfs_inode *ip = (struct gfs_inode *) file->f_dentry->d_inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	uint64_t start = fl->fl_start, end = fl->fl_end;
	long pid = fl->fl_pid;
	int plock = (fl->fl_flags & FL_POSIX);
	int flock = (fl->fl_flags & FL_FLOCK);
	int get, set, wait, ex, sh, un;
	int error = -ENOLCK;

	if (sdp->sd_args.ar_localflocks) {
		error = LOCK_USE_CLNT;
		goto out;
	}

	if ((ip->i_di.di_mode & (S_ISGID | S_IXGRP)) == S_ISGID)
		goto out;

	if (!flock && !plock)
		goto out;

	error = -EINVAL;

	get = (IS_GETLK(cmd)) ? TRUE : FALSE;
	set = (IS_SETLK(cmd)) ? TRUE : FALSE;
	wait = (IS_SETLKW(cmd)) ? TRUE : FALSE;

	if ((flock && (get || (!set && !wait))) ||
	    (plock && (!get && !set && !wait)))
		goto out;

	ex = (fl->fl_type == F_WRLCK) ? TRUE : FALSE;
	sh = (fl->fl_type == F_RDLCK) ? TRUE : FALSE;
	un = (fl->fl_type == F_UNLCK) ? TRUE : FALSE;

	if (!ex && !sh && !un)
		goto out;

	if (flock) {
		struct gfs_file *fp = gfs_get_file(file);
		GFS_ASSERT(fp,);

		if (un)
			error = gfs_funlock(fp);
		else
			error = gfs_flock(fp, ex, wait);
	} else {
		if (get) {
			error =
			    gfs_plock_get(ip, (unsigned long) fl->fl_owner,
					  &start, &end, &ex, &pid);
			if (error < 0)
				goto out;

			fl->fl_type = F_UNLCK;
			if (!error)
				goto out;

			fl->fl_start = start;
			fl->fl_end = end;
			fl->fl_pid = pid;
			fl->fl_type = (ex) ? F_WRLCK : F_RDLCK;

			error = 0;
		} else if (un)
			error =
			    gfs_punlock(ip, (unsigned long) fl->fl_owner, start,
					end);
		else
			error =
			    gfs_plock(ip, (unsigned long) fl->fl_owner, wait,
				      ex, start, end);
	}

      out:

	return error;
}

/**
 * gfs_sendfile - Send bytes to a file or socket
 * @in_file: The file to read from
 * @out_file: The file to write to
 * @count: The amount of data
 * @offset: The beginning file offset
 *
 * Outputs: offset - updated according to number of bytes read
 *
 * Returns: The number of bytes sent, -EXXX on failure
 */

static ssize_t
gfs_sendfile(struct file *in_file, struct file *out_file,
	     size_t count, loff_t * offset)
{
	struct gfs_inode *ip = (struct gfs_inode *) in_file->f_dentry->d_inode->u.generic_ip;
	struct gfs_holder ghs[2];
	unsigned int num_gh = 1;
	ssize_t retval;

	gfs_init_holder(ip->i_gl, LM_ST_SHARED, GL_ATIME | GL_DATA_B, &ghs[0]);

	if (out_file->f_dentry &&
	    out_file->f_dentry->d_inode &&
	    out_file->f_dentry->d_inode->i_sb ==
	    in_file->f_dentry->d_inode->i_sb) {
		gfs_init_holder(((struct gfs_inode *) out_file->f_dentry->
				 d_inode->u.generic_ip)->i_gl, LM_ST_EXCLUSIVE,
				GL_DATA_B, &ghs[1]);
		num_gh = 2;
	}

	retval = gfs_glock_nq_m_atime(num_gh, ghs);
	if (retval)
		goto out;

	if (gfs_is_jdata(ip))
		retval = -ENOSYS;
	else {
		read_descriptor_t desc;

		desc.written = 0;
		desc.count = count;
		desc.buf = (char *) out_file;
		desc.error = 0;
		do_generic_file_read(in_file, offset, &desc, file_send_actor,
				     0);

		retval = desc.written;
		if (!retval)
			retval = desc.error;
	}

	gfs_glock_dq_m(num_gh, ghs);

      out:
	gfs_uninit_holder(&ghs[0]);
	if (num_gh == 2)
		gfs_uninit_holder(&ghs[1]);

	return retval;
}

/**
 * gfs_vma_open - 
 * @vma:
 *
 */

void gfs_vma_open(struct vm_area_struct *vma)
{
	struct gfs_inode *ip = (struct gfs_inode *)vma->vm_file->f_dentry->d_inode->u.generic_ip;
	struct list_head *tmp, *head;
	struct gfs_vma_kludge *vk = NULL, *vk_new = NULL;

	if (!(vma->vm_flags & VM_MAYSHARE))
		return;

	for (;;) {
		spin_lock(&ip->i_arch.i_lock);

		for (head = &ip->i_arch.i_mmap_shared, tmp = head->next;
		     tmp != head;
		     tmp = tmp->next) {
			vk = list_entry(tmp, struct gfs_vma_kludge, vk_list);
			if (vk->vk_vma == vma)
				break;
		}

		if (tmp == head) {
			if (vk_new) {
				vk = vk_new;
				list_add(&vk->vk_list, head);
				vk_new = NULL;
			} else
				vk = NULL;
		}

		spin_unlock(&ip->i_arch.i_lock);

		if (vk) {
			if (vk_new)
				kfree(vk_new);
			break;
		}

		vk_new = gmalloc(sizeof(struct gfs_vma_kludge));
		vk_new->vk_vma = vma;
	}
}

/**
 * gfs_vma_close - 
 * @vma:
 *
 */

void gfs_vma_close(struct vm_area_struct *vma)
{
	struct gfs_inode *ip = (struct gfs_inode *)vma->vm_file->f_dentry->d_inode->u.generic_ip;
	struct list_head *tmp, *head;
	struct gfs_vma_kludge *vk = NULL;

	if (!(vma->vm_flags & VM_MAYSHARE))
		return;

	zap_page_range(vma, vma->vm_start, vma->vm_end - vma->vm_start);

	spin_lock(&ip->i_arch.i_lock);

	for (head = &ip->i_arch.i_mmap_shared, tmp = head->next;
	     tmp != head;
	     tmp = tmp->next) {
		vk = list_entry(tmp, struct gfs_vma_kludge, vk_list);
		if (vk->vk_vma == vma) {
			list_del(&vk->vk_list);
			break;
		}
	}

	spin_unlock(&ip->i_arch.i_lock);

	if (tmp != head)
		kfree(vk);
}

/**
 * gfs_shared_nopage - do safe locking on private mappings
 * @area - vm_area to pass through to filemap_nopage
 * @address - pass through to filemap_nopage
 * @no_share - pass through to filemap_nopage
 *
 * Returns: the new page
 */

static struct page *
gfs_shared_nopage(struct vm_area_struct *area,
		  unsigned long address, int no_share)
{
	struct gfs_inode *ip =
	    (struct gfs_inode *) area->vm_file->f_dentry->d_inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_alloc *al = NULL;
	struct page *result = NULL;
	unsigned int data_blocks, ind_blocks;
	unsigned int blocks_per_page;
	uint64_t lblock, dblock;
	uint32_t extlen;
	unsigned long index;
	unsigned int x;
	int dirty_page = FALSE;
	int alloc_required;
	int error;

	GFS_ASSERT_INODE(gfs_glock_is_locked_by_me(ip->i_gl), ip,);

	if (gfs_is_jdata(ip))
		return NULL;

	set_bit(GIF_PAGED, &ip->i_flags);
	set_bit(GIF_SW_PAGED, &ip->i_flags);

	index =
	    ((address - area->vm_start) >> PAGE_CACHE_SHIFT) + area->vm_pgoff;

	error =
	    gfs_write_alloc_required(ip, (uint64_t) index << PAGE_CACHE_SHIFT,
				     PAGE_CACHE_SIZE, &alloc_required);
	if (error)
		goto fail;

	if (!alloc_required) {
		result = filemap_nopage(area, address, no_share);
		if (result && result != NOPAGE_OOM) {
			GFS_ASSERT_INODE(index == result->index, ip,);
			GFS_ASSERT_INODE(Page_Uptodate(result), ip,);
		}
		return result;
	}

	gfs_write_calc_reserv(ip, PAGE_CACHE_SIZE, &data_blocks, &ind_blocks);

	al = gfs_get_alloc(ip);

	error =
	    gfs_get_quota_locks(ip, NO_QUOTA_CHANGE, NO_QUOTA_CHANGE,
				gfs_get_user_cred());
	if (error)
		goto fail_alloc;

	al->al_metablks = ind_blocks;
	al->al_datablks = data_blocks;
	al->al_trans = TRANS_NOPAGE;

	error = gfs_inplace_reserv(ip);
	if (error)
		goto fail_gunlock_q;

	error = gfs_check_quota(ip, ip->i_di.di_uid, ip->i_di.di_gid);
	if (error)
		goto fail_ipres;

	/*  Trans may require:
	   a dinode block, RG bitmaps to allocate from, indirect blocks, and a quota block  */

	error =
	    gfs_trans_begin(sdp, TRANS_NOPAGE,
			    1 + al->al_rgd->rd_ri.ri_length + ind_blocks, 1);
	if (error)
		goto fail_ipres;

	/*  Always unstuff the dinode.  */

	if (gfs_is_stuffed(ip)) {
		error = gfs_unstuff_dinode(ip, gfs_unstuffer_page, NULL);
		if (error)
			goto fail_end_trans;
	}

	/*  Fetch the page  */

	result = filemap_nopage(area, address, no_share);
	if (!result || result == NOPAGE_OOM)
		goto fail_end_trans;

	GFS_ASSERT_INODE(index == result->index, ip,);
	GFS_ASSERT_INODE(Page_Uptodate(result), ip,);

	/*  Figure out what blocks we need  */

	blocks_per_page = 1 << (PAGE_CACHE_SHIFT - sdp->sd_sb.sb_bsize_shift);
	lblock =
	    (uint64_t) index << (PAGE_CACHE_SHIFT - sdp->sd_sb.sb_bsize_shift);

	/*  Allocate blocks if we need to  */

	for (x = 0; x < blocks_per_page; x += extlen) {
		int new = 1;

		error = gfs_block_map(ip, lblock, &new, &dblock, &extlen);
		if (error)
			goto fail_new_page;
		GFS_ASSERT_INODE(dblock, ip,);

		if (new)
			dirty_page = TRUE;

		lblock += extlen;
	}

	if (dirty_page)
		set_page_dirty(result);

	gfs_trans_end(sdp);

	GFS_ASSERT_INODE(al->al_metaalloced || al->al_dataalloced, ip,);
	gfs_inplace_relse(ip);
	gfs_put_quota_locks(ip);
	gfs_put_alloc(ip);

	return result;

      fail_new_page:
	ClearPageUptodate(result);
	page_cache_release(result);
	result = NULL;

      fail_end_trans:
	gfs_trans_end(sdp);

      fail_ipres:
	gfs_inplace_relse(ip);

      fail_gunlock_q:
	gfs_put_quota_locks(ip);

      fail_alloc:
	gfs_put_alloc(ip);

      fail:
	return result;
}

/**
 * gfs_shared_do_no_page - Pull in a page that's part of a SWmmap
 * @mm - memory map of current
 * @area - vm_area to insert page into
 * @address - pass through to filemap_nopage
 * @write_access - set if triggered by a write fault
 * @page_table - where to write the pte
 *
 * Currently we don't support the shared state. Thus, if somebody owns a page
 * in the file, they also have an exclusive lock on the whole file.
 *
 * We enter with the page_table_lock held, and must leave with it held.
 *
 * Returns: 1 for minor fault, 2 for major fault, 0 for bus fault, -1 for OOM
 */

static int
gfs_shared_do_no_page(struct mm_struct *mm, struct vm_area_struct *area,
		      unsigned long address, int write_access,
		      pte_t * page_table, pmd_t * pmd)
{
	struct gfs_inode *ip =
	    (struct gfs_inode *) area->vm_file->f_dentry->d_inode->u.generic_ip;
	struct gfs_holder i_gh;
	int result;
	int error;

	pte_unmap(page_table);
	spin_unlock(&mm->page_table_lock);

	error = gfs_glock_nq_init(ip->i_gl, LM_ST_EXCLUSIVE, GL_DATA_B, &i_gh);
	if (error)
		return 0;

	spin_lock(&mm->page_table_lock);
	page_table = pte_offset_map(pmd, address);
	result = do_no_page(mm, area, address, write_access, page_table, pmd);

	gfs_glock_dq_uninit(&i_gh);

	return result;
}

static struct vm_operations_struct gfs_shared_mmap = {
      open:gfs_vma_open,
      close:gfs_vma_close,
      nopage:gfs_shared_nopage,
      do_no_page:gfs_shared_do_no_page
};

/**
 * gfs_private_nopage - do safe locking on private mappings
 * @area - vm_area to pass through to filemap_nopage
 * @address - pass through to filemap_nopage
 * @no_share - pass through to filemap_nopage
 *
 * Returns: the new page
 */

static struct page *
gfs_private_nopage(struct vm_area_struct *area,
		   unsigned long address, int no_share)
{
	struct gfs_inode *ip =
	    (struct gfs_inode *) area->vm_file->f_dentry->d_inode->u.generic_ip;
	struct gfs_holder i_gh;
	struct page *result;
	int error;

	error = gfs_glock_nq_init(ip->i_gl, LM_ST_SHARED, GL_DATA_B, &i_gh);
	if (error)
		return NULL;

	set_bit(GIF_PAGED, &ip->i_flags);

	result = filemap_nopage(area, address, no_share);

	gfs_glock_dq_uninit(&i_gh);

	return result;
}

static struct vm_operations_struct gfs_private_mmap = {
      open:gfs_vma_open,
      close:gfs_vma_close,
      nopage:gfs_private_nopage
};

/**
 * gfs_file_mmap - We don't support shared writable mappings right now
 * @file: The file to map
 * @vma: The VMA which described the mapping
 *
 * Returns: 0 or error code
 */

static int
gfs_file_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct gfs_inode *ip = (struct gfs_inode *) file->f_dentry->d_inode->u.generic_ip;
	struct gfs_holder i_gh;
	int error;

	gfs_init_holder(ip->i_gl, LM_ST_SHARED, GL_ATIME, &i_gh);
	error = gfs_glock_nq_atime(&i_gh);
	if (error) {
		gfs_uninit_holder(&i_gh);
		return error;
	}

	if (!gfs_is_jdata(ip)) {
		/* If this mapping isn't writeable at all, we don't need to acquire
		   the exclusive lock and we can just treat it as private mapping
		   (except for invalidation).  This is VM_MAYWRITE instead of
		   VM_WRITE because a call to mprotect() can turn on VM_WRITE
		   later. */

		if ((vma->vm_flags & (VM_MAYSHARE | VM_MAYWRITE)) == (VM_MAYSHARE | VM_MAYWRITE))
			vma->vm_ops = &gfs_shared_mmap;
		else
			vma->vm_ops = &gfs_private_mmap;
	} else {
		if (vma->vm_flags & VM_MAYSHARE)
			error = -ENOSYS;
		else
			vma->vm_ops = &gfs_private_mmap;
	}

	gfs_glock_dq_uninit(&i_gh);

	if (!error)
		gfs_vma_open(vma);

	return error;
}

struct file_operations_ext gfs_file_fops = {
      f_op_orig: {
	      llseek:gfs_llseek,
	      read:gfs_read,
	      write:gfs_write,
	      ioctl:gfs_ioctl,
	      mmap:gfs_file_mmap,
	      open:gfs_open,
	      release:gfs_close,
	      fsync:gfs_sync_file,
	      lock:gfs_lock_file
      },
      sendfile:gfs_sendfile,
};

struct file_operations gfs_dir_fops = {
      readdir:gfs_readdir,
      ioctl:gfs_ioctl,
      open:gfs_open,
      release:gfs_close,
      fsync:gfs_sync_file,
      lock:gfs_lock_file,
};
