/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#include <asm/pgalloc.h>
#include <linux/blkdev.h>

#include "bmap.h"
#include "dio.h"
#include "linux_inode.h"
#include "linux_page.h"

/**
 * unmap_shared_mapping - Mark dirty pages dirty and then kill the PTEs for a mapping
 * @mapping: The address space
 *
 */

static void
unmap_shared_mapping(struct address_space *mapping)
{
	struct gfs_inode *ip = (struct gfs_inode *)mapping->host->u.generic_ip;
	struct list_head *tmp, *head;
	struct gfs_vma_kludge *vk;
	struct vm_area_struct *vma;

	spin_lock(&ip->i_arch.i_lock);

	for (head = &ip->i_arch.i_mmap_shared, tmp = head->next;
	     tmp != head;
	     tmp = tmp->next) {
		vk = list_entry(tmp, struct gfs_vma_kludge, vk_list);
		vma = vk->vk_vma;
		zap_page_range(vma, vma->vm_start, vma->vm_end - vma->vm_start);
	}

	spin_unlock(&ip->i_arch.i_lock);
}

/**
 * gfs_inval_pte - Sync and invalidate all PTEs associated with a glock
 * @gl: the glock
 *
 */

void
gfs_inval_pte(struct gfs_glock *gl)
{
	struct gfs_inode *ip;
	struct inode *inode;

	ip = (struct gfs_inode *)gl->gl_object;
	if (!ip ||
	    ip->i_di.di_type != GFS_FILE_REG)
		return;

	if (!test_bit(GIF_PAGED, &ip->i_flags))
		return;

	inode = gfs_iget(ip, NO_ST_CREATE);
	if (inode) {
		unmap_shared_mapping(inode->i_mapping);
		iput(inode);

		if (test_bit(GIF_SW_PAGED, &ip->i_flags))
			set_bit(GLF_DIRTY, &gl->gl_flags);
	}

	clear_bit(GIF_SW_PAGED, &ip->i_flags);
}

/**
 * gfs_inval_page - Invalidate all pages associated with a glock
 * @gl: the glock
 *
 */

void
gfs_inval_page(struct gfs_glock *gl)
{
	struct gfs_inode *ip;
	struct inode *inode;

	ip = (struct gfs_inode *)gl->gl_object;
	if (!ip ||
	    ip->i_di.di_type != GFS_FILE_REG)
		return;

	inode = gfs_iget(ip, NO_ST_CREATE);
	if (inode) {
		struct address_space *mapping = inode->i_mapping;

		GFS_ASSERT_INODE(list_empty(&inode->i_dirty_buffers) &&
				 list_empty(&inode->i_dirty_data_buffers),
				 ip,);
		GFS_ASSERT_INODE(list_empty(&mapping->dirty_pages) &&
				 list_empty(&mapping->locked_pages),
				 ip,);

		truncate_inode_pages(mapping, 0);

		GFS_ASSERT_INODE(!mapping->nrpages, ip,);

		iput(inode);
	}

	clear_bit(GIF_PAGED, &ip->i_flags);
}

/**
 * wait_on_mapping - 
 * @mapping:
 *
 * I wish we didn't have to do this.
 * filemap_fdata[sync|wait]() don't guarantee all pages are
 * written if a page cleaner decides to call them at the same
 * time as we do.
 *
 */

#define page_busy(page) ((page)->flags & ((1ul << PG_dirty) | (1ul << PG_locked)))

static void
wait_on_mapping(struct address_space *mapping)
{
	struct list_head *tmp, *head;
	struct page *page;
	int error;

 restart:
	lock_pagecache();

	for (head = &mapping->clean_pages, tmp = head->next;
	     tmp != head;
	     tmp = tmp->next) {
		page = list_entry(tmp, struct page, list);

		if (!page_busy(page))
			continue;

		page_cache_get(page);
		unlock_pagecache();
      
		lock_page(page);
		
		if (PageDirty(page)) {
			ClearPageDirty(page);
			error = mapping->a_ops->writepage(page);
			GFS_ASSERT(!error, );
		} else
			UnlockPage(page);

		page_cache_release(page);

		goto restart;
	}

	unlock_pagecache();
}

/**
 * gfs_sync_page_i - Sync the pages for a struct inode
 * @inode: the inode
 * @flags:
 *
 */

void
gfs_sync_page_i(struct inode *inode, int flags)
{
	struct address_space *mapping = inode->i_mapping;
	struct gfs_inode *ip = (struct gfs_inode *)inode->u.generic_ip;
	struct inode *aspace = ip->i_gl->gl_arch.gl_aspace;
	int error;

	error = filemap_fdatasync(mapping);
	GFS_ASSERT(!error, );

	down(&aspace->i_sem);
	fsync_inode_buffers(inode);
	fsync_inode_data_buffers(inode);
	up(&aspace->i_sem);

	error = filemap_fdatawait(mapping);
	GFS_ASSERT(!error, );

	if (test_bit(GIF_PAGED, &ip->i_flags))
		wait_on_mapping(mapping);

	if ((flags & (DIO_INVISIBLE | DIO_CHECK)) == DIO_CHECK) {
		GFS_ASSERT_INODE(list_empty(&inode->i_dirty_buffers) &&
				 list_empty(&inode->i_dirty_data_buffers),
				 ip,);
		GFS_ASSERT_INODE(list_empty(&mapping->dirty_pages) &&
				 list_empty(&mapping->locked_pages),
				 ip,);
	}
}

/**
 * gfs_sync_page - sync the pages associated with a glock
 * @gl: the glock
 * @flags:
 *
 */

void
gfs_sync_page(struct gfs_glock *gl, int flags)
{
	struct gfs_inode *ip;
	struct inode *inode;

	ip = (struct gfs_inode *)gl->gl_object;
	if (!ip ||
	    ip->i_di.di_type != GFS_FILE_REG)
		return;

	inode = gfs_iget(ip, NO_ST_CREATE);
	if (inode) {
		gfs_sync_page_i(inode, flags);
		iput(inode);
	}
}

/**
 * gfs_unstuffer_page - unstuff a stuffed inode into a block cached by a page
 * @ip: the inode
 * @dibh: the dinode buffer
 * @block: the block number that was allocated
 * @private: any locked page held by the caller process
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
gfs_unstuffer_page(struct gfs_inode *ip,
		   struct buffer_head *dibh, uint64_t block, void *private)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct inode *inode = ip->i_vnode;
	struct page *page = (struct page *) private;
	struct buffer_head *bh;
	void *kaddr;
	int release = FALSE;
	int error = 0;

	GFS_ASSERT_INODE(inode, ip,);

	if (!page || page->index) {
		RETRY_MALLOC(page = grab_cache_page(inode->i_mapping, 0), page);
		release = TRUE;
	}

	if (!Page_Uptodate(page)) {
		kaddr = kmap(page);
		memcpy(kaddr,
		       (dibh)->b_data + sizeof (struct gfs_dinode),
		       ip->i_di.di_size);
		memset(kaddr + ip->i_di.di_size, 0,
		       PAGE_CACHE_SIZE - ip->i_di.di_size);
		kunmap(page);

		SetPageUptodate(page);
	}

	if (!page->buffers)
		create_empty_buffers(page, sdp->sd_vfs->s_dev,
				     sdp->sd_sb.sb_bsize);

	bh = page->buffers;
	GFS_ASSERT_INODE(bh, ip,);
	GFS_ASSERT_INODE(!buffer_mapped(bh), ip,);

	bh->b_dev = sdp->sd_vfs->s_dev;
	bh->b_blocknr = block;
	bh->b_state |= 1UL << BH_Mapped;
	mark_buffer_uptodate(bh, TRUE);
	mark_buffer_dirty(bh);
	buffer_insert_inode_data_queue(bh, inode);

	if (release) {
		UnlockPage(page);
		page_cache_release(page);
	}

	return error;
}

/**
 * gfs_truncator_page - truncate a partial data block in the page cache
 * @ip: the inode
 * @size: the size the file should be
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
gfs_truncator_page(struct gfs_inode *ip, uint64_t size)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct inode *inode = ip->i_vnode;
	struct page *page;
	struct buffer_head *bh;
	void *kaddr;
	uint64_t bn;
	unsigned long index;
	unsigned int offset, bufnum;
	int not_new = 0;
	int error = 0;

	GFS_ASSERT_INODE(inode, ip,);

	error =
	    gfs_block_map(ip, size >> sdp->sd_sb.sb_bsize_shift, &not_new, &bn,
			  NULL);
	if (error)
		goto out;

	if (!bn)
		goto out;

	index = size >> PAGE_CACHE_SHIFT;
	offset = size & (PAGE_CACHE_SIZE - 1);
	bufnum = offset / sdp->sd_sb.sb_bsize;

	/*  Not in a transaction here -- a non-disk-I/O error is ok.  */

	page =
	    read_cache_page(inode->i_mapping, index,
			    (filler_t *) inode->i_mapping->a_ops->readpage,
			    NULL);
	if (IS_ERR(page)) {
		error = PTR_ERR(page);
		goto out;
	}

	lock_page(page);

	if (!Page_Uptodate(page) || PageError(page)) {
		error = -EIO;
		goto out_release;
	}

	kaddr = kmap(page);
	memset(kaddr + offset, 0, PAGE_CACHE_SIZE - offset);
	kunmap(page);

	if (!page->buffers) {
		create_empty_buffers(page, sdp->sd_vfs->s_dev,
				     sdp->sd_sb.sb_bsize);
		GFS_ASSERT_INODE(page->buffers, ip,);
	}

	for (bh = page->buffers; bufnum--; bh = bh->b_this_page)	/* Do nothing */
		;

	if (!buffer_mapped(bh)) {
		bh->b_dev = sdp->sd_vfs->s_dev;
		bh->b_blocknr = bn;
		bh->b_state |= 1UL << BH_Mapped;
		mark_buffer_uptodate(bh, TRUE);
	} else
		GFS_ASSERT_INODE(bh->b_dev == sdp->sd_vfs->s_dev
				 && bh->b_blocknr == bn, ip,
				 printk("0x%.4X 0x%.4X %" PRIu64 " %" PRIu64
					"\n", (unsigned int) bh->b_dev,
					(unsigned int) sdp->sd_vfs->s_dev,
					(uint64_t) bh->b_blocknr, bn);
	    );

	mark_buffer_dirty(bh);
	buffer_insert_inode_data_queue(bh, inode);

      out_release:
	UnlockPage(page);
	page_cache_release(page);

      out:
	return error;
}
