/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "dio.h"
#include "glock.h"
#include "locking.h"
#include "lvb.h"
#include "super.h"

/**
 * gfs_mount_lockproto - mount a locking protocol
 * @sdp: the filesystem
 * @args: mount arguements
 * @silent: if TRUE, don't complain if the FS isn't a GFS fs
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
gfs_mount_lockproto(struct gfs_sbd *sdp, int silent)
{
	struct gfs_sb sb;
	struct buffer_head *bh;
	char proto[256], table[256], host[256];
	int error = 0;

	/*  Copy in values from "args"  */

	strcpy(proto, sdp->sd_args.ar_lockproto);
	strcpy(table, sdp->sd_args.ar_locktable);
	strcpy(host, sdp->sd_args.ar_hostdata);

	/*  Try to autodetect  */

	if (!proto[0] || !table[0]) {
		error =
		    gfs_dread(sdp, GFS_SB_ADDR >> sdp->sd_fsb2bb_shift, NULL,
			      DIO_FORCE | DIO_START | DIO_WAIT, &bh);
		if (error)
			goto out;
		gfs_sb_in(&sb, (bh)->b_data);
		gfs_drelse(bh);

		error = gfs_check_sb(sdp, &sb, silent);
		if (error)
			goto out;

		if (!proto[0])
			strcpy(proto, sb.sb_lockproto);

		if (!table[0])
			strcpy(table, sb.sb_locktable);
	}

	error = lm_mount(proto, table, host,
			 gfs_glock_cb, sdp,
			 GFS_MIN_LVB_SIZE, &sdp->sd_lockstruct);
	if (error) {
		printk
		    ("GFS: can't mount proto = %s, table = %s, hostdata = %s\n",
		     proto, table, host);
		goto out;
	}

	GFS_ASSERT_SBD(sdp->sd_lockstruct.ls_lockspace, sdp,);
	GFS_ASSERT_SBD(sdp->sd_lockstruct.ls_ops, sdp,);
	GFS_ASSERT_SBD(sdp->sd_lockstruct.ls_lvb_size >= GFS_MIN_LVB_SIZE,
		       sdp,);

	if (!*table)
		gfs_device_name(sdp, table, 256);

	snprintf(sdp->sd_fsname, 256, "%s.%u", table,
		 sdp->sd_lockstruct.ls_jid);

      out:
	return error;
}

/**
 * gfs_unmount_lockproto - Unmount lock protocol
 * @sdp: The GFS superblock
 *
 */

void
gfs_unmount_lockproto(struct gfs_sbd *sdp)
{
	lm_unmount(&sdp->sd_lockstruct);
}
