%{
/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "sisstring.h"
#include "sislogger.h"
#include "osi_user.h"
#include "pool_std.h"
#include "libpool.h"
#include "libpool_internal.h"

/* FIXME: I still need to have more error checking in the parsing */
/* FIXME: Make sure that pool name is not too long */

user_pool_t *pool = NULL;
int minor_set = 0;
uint32 sp_count = 0;
uint32 currentspid = 0;
uint32 dev_count = 0;
extern FILE *yyin;

int _set_subpools(uint32 num);
int _add_subpool(uint32 id, uint32 stripe, uint32 numdev, uint32 type);
int _add_label(uint32 spid, uint32 devid, char *devname);

/* FIXME: There has to be a better way to prevent compiler from complaining*/
size_t strnlen(const char *s, size_t maxlen);
int yyerror(char *s);
int yylex();

%}

%union {
   int num;
   char *str;
}

%token POOLNAME MINOR SUBPOOLS SUBPOOL POOLDEV
%token <num> NUMBER
%token <str> STRING DEVICESTR

%type <num> minorcheck spcount sp spstripe spdevs sptype
%type <num> devnum devweight
%type <str> poolcheck devname
%%
lines: line
  | lines line
  ;

line: POOLNAME poolcheck {
         if(strnlen(pool->name, POOL_NAME_SIZE) > 0) {
            yyerror("Multiple pool names defined in config file.");
            return 1;
         }
         else {
            strncpy(pool->name, $2, POOL_NAME_SIZE);
            pool->name[POOL_NAME_SIZE-1] = '\0';
         }
         free($2);
      }
  | MINOR minorcheck  {
     if(minor_set) {
        yyerror("Minor number multiply defined.");
        return 1;
     }
     pool->minor = $2;
     minor_set = 1;
  }
  | SUBPOOLS spcount { if(_set_subpools($2))
     return 1; }
  | SUBPOOL sp spstripe spdevs sptype { if(_add_subpool($2, $3, $4, $5))
     return 1; }

  | POOLDEV sp devnum devname devweight {   if(_add_label($2, $3, $4))
     return 1; }
  | '\n'  { /* ignore extraneous newlines */ }
;

poolcheck: STRING {
              char msg[256];
              if(check_poolname($1) != 0) {
                 snprintf(msg,256, "Pool name must be less than %u characters",
                          (POOL_NAME_SIZE - 1));
                 yyerror(msg);
                 return 1;
              }
              $$ = $1;
           }
  | DEVICESTR {
        yyerror("Pool name cannot have a '/' in it");
        return 1;
  }
  | NUMBER {
          char *name = malloc(sizeof(char) * 256);
          memset(name, '\0', 256);
          snprintf(name, 256, "%d", $1);
          if(name[255] != '\0') {
                  name[255] = '\0';
                  /* I don't expect to see this happen... */
                  yyerror("Truncated poolname");
          }
          $$ = name;
  }
;

minorcheck: NUMBER {
            if(check_minornum($1) != 0) {
               yyerror("Minor number must be less than 65.\n"
                     "Use 0 to allow pool to choose a minor number.");
               return 1;
            }
            $$ = $1;
         }
;

spcount: NUMBER {
            if(pool->num_sps != 0) {
               yyerror("Number of subpools multiply defined.");
               return 1;
            }
            if($1 == 0) {
               yyerror("Number of subpools must be greater than zero.");
               return 1;
            }
            $$ = $1;
   }


;

sp:  NUMBER {
        if($1 >= pool->num_sps) {
           yyerror("Subpool ID out of bounds.\n"
                 "Make sure the number of subpools defined in your"
                 " config is correct");
           return 1;
        }
        currentspid = $1;
        $$ = $1;
     }
;

spstripe: NUMBER {
              /* FIXME: Using signed integers and atoi to get numbers
               * out of the config file right now - should be using
               * unsigned ints and strtoul.
               * If we switch to this, this needs to be UINT_MAX */
             if($1 >= INT_MAX) {
                yyerror("Subpool stripe size out of bounds");
                return 1;
             }
             /* FIXME: Assuming 512 byte sectors */
             if((($1 >> 3) << 3) != $1) {
                yyerror("Subpool stripe size must be a multiple of 8 (4kB)");
                return 1;
             }
             $$ = $1;}
;

spdevs: NUMBER {
           char msg[500];
           if($1 > POOL_MAX_DEVICES) {
              snprintf(msg, 500, "Max number of devices per subpool (%u)"
                    " exceeded", POOL_MAX_DEVICES);
              yyerror(msg);
              return 1;
           }
           $$ = $1;
        }
;

sptype: STRING {
           uint32 tmp;
           if(get_sptype_num($1, &tmp) != 0) {
              yyerror("Invalid subpool type name");
              return 1;
           }
           $$ = tmp;
        }
    | {
       uint32 tmp;
       if(get_sptype_num("gfs_data", &tmp)) {
           yyerror("Invalid subpool type name");
           return 1;
       }
       $$ = tmp;
      }
;

devnum: NUMBER {
           char msg[256];
           if($1 >= pool->sp[currentspid].num_devs) {
              snprintf(msg, 256, "Device ID out of bounds.\n"
                       "Make sure the number of devices defined for subpool %u"
                       " is correct", currentspid);
              yyerror(msg);
              return 1;
           }
           if(pool->sp[currentspid].devs[$1].initialized > 0) {
                   snprintf(msg, 256, "Device ID %d already defined"
                            " for subpool %u", $1, currentspid);
                   yyerror(msg);
                   return 1;
           }
           $$ = $1;
        }
;

devname: DEVICESTR {
           int i, j;
           struct stat s;
           char msg[256];

           /* scan all the previous devices in the pool */
           for(i = 0; i <= currentspid; i++) {
              for(j = 0; j < pool->sp[i].num_devs; j++) {
                 if(pool->sp[i].devs != NULL &&
                    pool->sp[i].devs[j].dev_path[0] != NULL) {
                    if(strcmp($1, pool->sp[i].devs[j].dev_path[0]) == 0) {
                       yyerror("Duplicate device specified");
                       return 1;
                    }
                 }
              }
           }
           if(stat($1, &s)) {
                   snprintf(msg, 256, "Unable to get file status: %s",
                            strerror(errno));
                   yyerror(msg);
                   return 1;
           }
           if(!S_ISBLK(s.st_mode)) {
                   snprintf(msg, 256, "%s is not a block device", $1);
                   yyerror(msg);
                   return 1;
           }
           $$ = $1;
         }
   | STRING {
        yyerror("Device name invalid");
        return 1;
   }
;

devweight: NUMBER {
              yyerror("Warning: Hardware DMEP weight deprecated and will be ignored.");
              $$ = 0;
           }
         | {
            $$ = 0;
         }
;

%%


int lineno = 1;


/** Fxns Internal to libpool **/

/**
 * check_poolname - verifies poolname conforms to allowed parameters
 * @pn: name of pool
 *
 * Description: Verifies poolname is not too large
 *
 * Returns: 0 on success, non-zero on error
 */
int check_poolname(char *pn) {
   if(strlen(pn) > (POOL_NAME_SIZE - 1)) {
      return 1;
   }
   else {
      return 0;
   }
}


/**
 * check_minornum - verifies minor number conforms to allowed parameters
 * @minor - minor number to check
 *
 * Description: verifies minor number is in bounds
 *
 * Returns: 0 on success, non-zero on error
 */
int check_minornum(uint32 minor) {
   if(minor > MAX_HARDCODED_MINOR) {
      return 1;
   }
   else {
      return 0;
   }
}

/** Fxns Internal to pool_config_grammar.y **/

/**
 * _check_pool - verifies that pool is valid
 *
 * Description: checks that the correct number of subpools and devices
 *   are in the pool structure
 *
 * FIXME: add code here ;)
 * Returns: 0 on success, non-zero on error.
 */
int _check_pool(void)
{
   int error = 0;
   int i, j;
   char msg[256];

   for(i = 0; i < pool->num_sps; i++) {
      if(pool->sp[i].num_devs == 0) {
         snprintf(msg, 256, "%d subpools specified - subpool %d missing.",
                  pool->num_sps, i);
         yyerror(msg);
         error = 1;
      }
   }

   for(i = 0; i < pool->num_sps; i++) {
      for(j = 0; j < pool->sp[i].num_devs; j++) {
         if(pool->sp[i].devs[j].initialized == 0) {
            snprintf(msg, 256, "subpool %d specified %d pool devices - "
                     "%d defined.",
                    i, pool->sp[i].num_devs, j);
            yyerror(msg);
            error = 1;
         }
      }
   }

   return error;
}

/**
 * _parse_pool_config -
 * @fp:
 * @p:
 *
 * Returns: 0 on success, non-zero on error.
 */
int _parse_pool_config(FILE *fp, user_pool_t **p)
{
   int error = 0;

   yyin = fp;

   pool = (user_pool_t *) malloc(sizeof(user_pool_t));

   if(pool == NULL) {
      yyerror("Out of memory!");
      error = -1;
      goto end;
   }
   memset(pool, 0, sizeof(user_pool_t));

   if((error = yyparse()) != 0) {
      free(pool);
      pool = NULL;
      goto end;
   }

   if((error = _check_pool()) != 0) {
      free(pool);
      pool = NULL;
      goto end;
   }

end:
   *p = pool;
   return error;
}

/**
 * _set_subpools - setup the subpools
 * @num: number of subpools
 *
 * sets the number of subpools in the pool and allocates memory for their
 * fields.
 *
 * Returns: zero on success, non-zero on failure
 */
int _set_subpools(uint32 num)
{

   pool->num_sps = num;
   /* we need a subpool structure for each sp */
   pool->sp = (user_subpool_t*) malloc(sizeof(user_subpool_t)* pool->num_sps);
   if(!pool->sp) {
      yyerror("Out of Memory");
      return -1;
   }
   memset(pool->sp, 0, sizeof(user_subpool_t) * pool->num_sps);
   return 0;
}

/**
 * _add_subpool -
 *
 */
int _add_subpool(uint32 id, uint32 striping, uint32 numdevs, uint32 type)
{
   user_subpool_t *subs;
   char msg[256];
   if(id >= pool->num_sps) {
      yyerror("Subpool ID out of bounds.");
      return 1;
   }
   subs = &(pool->sp[id]);
   if(subs->num_devs != 0) {
      snprintf(msg, 256, "Subpool %u already defined", id);
      yyerror(msg);
      return 1;
   }
   if(numdevs == 0) {
      yyerror("Number of devices must be greater than 0");
      return 1;
   }
	subs->id = id;
	subs->striping = striping;
	subs->num_devs = numdevs;
   subs->type = type;
	sp_count++;
   subs->devs = (user_device_t *)
      malloc(sizeof(user_device_t) * subs->num_devs);
   if(!subs->devs) {
      yyerror("Out of memory");
      return 1;
   }
   memset(subs->devs, 0, sizeof(user_device_t) * subs->num_devs);
   return 0;
}

/**
 * _add_label -
 *
 */
int _add_label(uint32 spid, uint32 devid, char *devname)
{

   user_subpool_t *subs = &pool->sp[spid];
   user_device_t *dev;
   if(!subs->devs) {
      yyerror("Subpool for device not initialized.");
      return 1;
   }
   if(devid >= subs->num_devs) {
      yyerror("Too many devices for subpool");
      return 1;
   }
   dev = &(subs->devs[devid]);

   if(dev->initialized) {
      yyerror("Device ID multiply defined for subpool.");
      return 1;
   }

   dev->sp_id = spid;
   dev->devid = devid;
   dev->path_count = 1;
   dev->good_path_count = 1;
   dev->dev_path[0] = (char *) malloc(sizeof(char) * strlen(devname));
   if(!dev->dev_path[0]) {
      yyerror("Out of memory");
      return 1;
   }
   strcpy(dev->dev_path[0], devname);
   dev->initialized = 1;
   dev_count++;

   return 0;
}

/**
 * _init_parser - (re)set parser variables
 *
 * Description: allows the parser to be run multiple times in a single
 * instantiation of the code
 */
void _init_parser(void) {
   pool = NULL;
   minor_set = 0;
   sp_count = 0;
   currentspid = 0;
   dev_count = 0;
}

/**
 * read_pool_config -
 * @cfg_file: name of config file to read
 * @p: structure to fill with config file information
 *
 * Description: This is the interface to the pool config file parser
 */
int read_pool_config(char *cfg_file, user_pool_t **p)
{
   int error = 0;
   FILE *cfgfd = NULL;
   user_pool_t *pl = NULL;
   char rp[PATH_MAX];
   struct stat s;

   _init_parser();

   if(cfg_file) {
      if(!realpath(cfg_file, rp)) {
         log_err("%s when reading from %s", strerror(errno), cfg_file);
         return 1;
      }

      if(stat(rp, &s)) {
         log_err("Unable to get file status for %s: %s", cfg_file,
                 strerror(errno));
         return 1;
      }
      if(!S_ISREG(s.st_mode)) {
         log_err("%s does not resolve to a regular file", cfg_file);
         return 1;
      }
      if(!(cfgfd = fopen(cfg_file, "r"))) {
         log_err("Unable to open %s: %s", cfg_file, strerror(errno));
      }
   }
   else {
      if(!(cfgfd = stdin)) {
         log_err("Unable to get stdin");
         return 1;
      }
   }

   error = _parse_pool_config(cfgfd, &pl);

   *p = pl;

   return error;
}

/**
 * print_config - reproduce the configuration file
 * pool -- a pointer to the pool structure
 *
 * attempts to reproduce the config file based on the pool strucutre
 */
int print_config(user_pool_t *pool)
{
   int i, j;
   user_subpool_t *sub;
   user_device_t *dev;
   char *name = NULL;
   printf("poolname %s\n", pool->name);
   if(pool->minor == 0 || pool->minor > 64) {
      printf("#minor <dynamically assigned>\n");
   }
   else {
      printf("minor %d\n", pool->minor);
   }
   printf("subpools %d\n", pool->num_sps);

   for(i = 0; i < pool->num_sps; i++) {
      sub = &pool->sp[i];
      printf("subpool %d ", sub->id);
      printf("%u ", sub->striping);
      printf("%u ", sub->num_devs);
      get_sptype_name(sub->type, &name);
      printf("%s\n", name == NULL ? "invalid" : name);
   }
   for(i = 0; i < pool->num_sps; i++) {
      sub = &pool->sp[i];

      for(j = 0; j < sub->num_devs; j++) {
         dev = &sub->devs[j];
         printf("pooldevice %d ", dev->sp_id);
         printf("%u ", dev->devid);
         if(dev->dev_path[0])
            printf("%s\n", dev->dev_path[0]);
         else
            printf("DEVICE_NOT_FOUND\n");
      }
   }
   return 0;
}

/*
 * vim: et ts=3 sw=3 noai cin :
 */

