/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

/* pool_info.c - Queries kernel regarding pool state */

#include "sisstring.h"
#include "siscli.h"
#include "sislogger.h"
#include "libpool.h"
#include "osi_user.h"
#include "pool_release.cf"
#include "copyright.cf"

#define STRING_SIZE 512

static void _print_info_label(struct diskinfo *di, int pcount,
                              char **poolnames, int *maxname);
static void _print_stats(char *poolname, user_poolstat_t *iostat);
static int _print_verbose_info(struct user_pool *pl, int hr,
                               struct sstring *emsg);
static void _print_info(user_pool_t *pl, int hr, int len);

static void _usage(char *name) {

   printf("Usage:\n");
   printf("\n");
   printf("%s [options] [pool(s)]\n", name);
   printf("\n");
   printf("Options:\n");
   printf("  -c               Clear all pool statistics.\n");
   printf("  -D               Enable debugging output.\n");
   printf("  -H               Display capacity in human readable format.\n");
   printf("  -h               This usage information.\n");
   printf("  -i               Pool information.\n");
   printf("  -p               Print configuration.\n");
   printf("  -q               Enable quiet output.\n");
   printf("  -s               Print pool statistics.\n");
   printf("  -t <sec>         Use with -s for continuous updates at <sec> "
          " intervals.\n");
   printf("  -V               Version information.\n");
   printf("  -v               Enable verbose output.\n");

   return;
}

int main(int argc, char **argv)
{
   int        error = 0,
              c,
              i,
              j,
              interval = 0,
              human_readable = 0,
              quiet = 0,
              verbose = 0;
   int        len = 0;
   int        prn_hdr = 0;
   int        namelen = 0;
   char       action = 0,
              **pool_names = NULL,
              msg[STRING_SIZE];
   diskinfo_t di;
   user_poolstat_t *pstat;
   struct sstring *emsg = NULL;

   while((c = getopt(argc, argv, "cDHhipqst:Vv")) != -1 && !error) {
      switch(c) {


         case 'c':
            if(action) {
               mutually_exclusive(argv[0], c, action);
               error = 1;
            }
            else
               action = c;
            break;

         case 'D':
            add_flag(POOL_DEBUG);
            verbose = 4;
            break;

         case 'H':
            human_readable = 1;
            break;

         case 'h':
            _usage(argv[0]);
            return error;
            break;

         case 'i':
            if(action) {
               mutually_exclusive(argv[0], c, action);
               error = 1;
            }
            else
               action = c;
            break;

         case 'p':
            if(action) {
               mutually_exclusive(argv[0], c, action);
               error = 1;
            }
            else
               action = c;
            break;

         case 'q':
            if(check_flag(POOL_VERBOSE)) {
               mutually_exclusive(argv[0], c, 'v');
               error = 1;
            }
            else {
               add_flag(POOL_QUIET);
               quiet++;
            }
            break;

         case 's':
            if(action) {
               mutually_exclusive(argv[0], c, action);
               error = 1;
            }
            else
               action = c;
            break;

         case 't':
            interval = atoi(optarg);
            break;

         case 'V':
            printf("%s %s (built %s %s)\n", argv[0],
                  POOL_RELEASE_NAME, __DATE__, __TIME__);
            printf("%s\n", GFS_COPYRIGHT);
            return error;
            break;

         case 'v':
            if(check_flag(POOL_QUIET)) {
               mutually_exclusive(argv[0], c, 'q');
               error = 1;
            }
            else {
               verbose++;
               add_flag(POOL_VERBOSE);
               break;
            }

         case ':':
         case '?':
            fprintf(stderr, "Please use '-h' for usage.\n");
            exit(1);
            break;

         default:
            fprintf(stderr, "Bad programmer! You forgot to catch the"
                 " %c flag\n", c);
            exit(1);
            break;
      }

   }

   if(argc > optind) {
      pool_names = &(argv[optind]);
      argc = argc - optind;
   }

   if(!error) {
      if((error = ccs_initlog("pool.ccs", "pool/tool_config",
                              verbose, quiet))) {
         if((error = init_defaultlog(verbose, quiet, NULL)) != 0) {
            fprintf(stderr, "Unable to initialize tool logger");
            goto end;
         }
      }

      emsg = ss_init(STRING_SIZE-1);

      if((error = open_comm(emsg))) {
         if(error == ENXIO) {
            log_err("Unable to open pool comm device.  "
                      "Verify pool module is loaded");
            cstring2ss(emsg, "", 0);
         }
      }
      else if ((error = check_version(emsg)) != 0) {
         ss2cstring(msg, STRING_SIZE, emsg);
         log_err(msg);
      }
      else {
         switch(action) {

            /* Info */
            case 0:
            case 'i':
               error = get_loaded_pools(&di, emsg);
               if(!error) {
                  if(pool_names != NULL) {
                     prn_hdr = 0;
                     for(j = 0; j < argc; j++) {
                        for(i = 0; i < di.pl_count; i++) {
                           if(!strcmp(pool_names[j], di.pl[i].name))
                              break;
                        }
                        if(i != di.pl_count) {
                           if(verbose > 0) {
                              _print_verbose_info(&(di.pl[i]), human_readable,
                                                   emsg);
                           }
                           else {
                              if(prn_hdr == 0) {
                                 _print_info_label(&di, argc, pool_names,
                                                   &namelen);

                                 prn_hdr = 1;
                              }
                              _print_info(&(di.pl[i]), human_readable,
                                          namelen);
                           }
                        }
                        else {
                           log_warn("No pool named %s found", pool_names[j]);
                           error = 1;
                        }
                     }
                  }
                  else {
                     if(di.pl_count != 0) {
                        if(verbose > 0) {
                           for(i = 0; i < di.pl_count; i++) {
                              _print_verbose_info(&(di.pl[i]),
                                                   human_readable, emsg);
                           }
                        }
                        else {
                           _print_info_label(&di, argc, pool_names,
                                             &namelen);
                           for(i = 0; i < di.pl_count; i++) {
                              _print_info(&(di.pl[i]), human_readable,
                                          namelen);
                           }
                        }
                     }
                     else {
                        log_warn("No pools found");
                        /*error = 1;*/
                     }
                  }
               }

               break;

            /* Clear stats */
            case 'c':

               if((error = get_loaded_pools(&di, emsg)) == 0) {
                  if(pool_names != NULL) {
                     for(j = 0; j < argc; j++) {
                        for(i = 0; i < di.pl_count; i++) {
                           if(strcmp(pool_names[j], di.pl[i].name) == 0)
                              break;
                        }
                        if(i < di.pl_count) {
                           clear_pool_stats(pool_names[j], emsg);
                           log_warn("%s statistics cleared", pool_names[j]);
                        }
                        else {
                           log_warn("No pool named %s found", pool_names[j]);
                           error |= 1;
                        }
                     }
                  }
                  else {
                     if(di.pl_count <= 0) {
                        log_warn("No pools currently assembled");
                        error = 1;
                        break;
                     }
                     else {
                        for(i = 0; i < di.pl_count; i++) {
                           clear_pool_stats(di.pl[i].name, emsg);
                           log_warn("%s statistics cleared", di.pl[i].name);
                        }
                     }
                  }
               }

               break;

            /* Print config */
            case 'p':
               if((error = get_loaded_pools(&di, emsg)) == 0) {
                  if(pool_names != NULL) {
                     prn_hdr = 0;
                     for(j = 0; j < argc; j++) {
                        for(i = 0; i < di.pl_count; i++) {
                           if(!(strcmp(pool_names[j], di.pl[i].name)))
                              break;
                        }
                        if(i < di.pl_count) {
                           if(argc > 1) {
                              if(prn_hdr == 0) {
                                 log_warn("# Break the following into separate"
                                          " files");
                                 prn_hdr = 1;
                              }
                              printf("#--- begin %s --------------------\n",
                                     di.pl[i].name);
                              print_config(&di.pl[i]);
                              printf("#--- end %s ------------------\n",
                                     di.pl[i].name);
                           }
                           else {
                              print_config(&di.pl[i]);
                           }
                        }
                        else {
                           log_err("No pool named %s found", pool_names[j]);
                           error |= 1;
                        }
                     }
                  }
                  else {
                     if(di.pl_count <= 0) {
                        ss_sprintf(emsg, "No pools currently assembled");
                        error = 1;
                     }
                     else {
                        if(di.pl_count == 1) {
                           print_config(&di.pl[0]);
                        }
                        else {
                           log_warn("# Break the following into separate"
                                  " files");
                           for(i = 0; i < di.pl_count; i++) {
                              printf("#--- begin %s --------------------\n",
                                     di.pl[i].name);
                              print_config(&di.pl[i]);
                              printf("#--- end %s ------------------\n",
                                     di.pl[i].name);
                           }
                        }
                     }
                  }
               }

               break;

            /* Stats */
            case 's':
               error = get_loaded_pools(&di, emsg);
               if(di.pl_count <= 0) {
                  log_warn("No pools currently assembled");
                  error = 1;
                  break;
               }

               if(!error) {
                  do {
                     if(pool_names != NULL) {
                        for(j = 0; j < argc; j++) {
                           for(i = 0; i < di.pl_count; i++) {
                              if(!(strcmp(pool_names[j], di.pl[i].name)))
                                 break;
                           }
                           if(i < di.pl_count) {
                              get_pool_stats(pool_names[j], &pstat, emsg);
                              _print_stats(pool_names[j], pstat);
                              destroy_pool_stats(&pstat);
                           }
                           else {
                              log_err("No pool named %s found",
                                         pool_names[j]);
                              error = 1;
                              continue;
                           }
                        }
                     }
                     else {
                        for(i = 0; i < di.pl_count; i++) {
                           get_pool_stats(di.pl[i].name, &pstat, emsg);
                           _print_stats(di.pl[i].name, pstat);
                           destroy_pool_stats(&pstat);
                        }
                     }
                  } while(interval && !sleep(interval));
               }

               break;

            default:
               log_err("Please use -h' for usage.");
               error = 1;
               break;
         }
      }
   }
   ss_len(emsg, &len);
   if(error != 0 && len > 0) {
      ss2cstring(msg, STRING_SIZE, emsg);
      log_err(msg);
   }
  end:
   close_comm(emsg);
   /* Until we have well defined error codes, the exit status will either
      be 0 or 1 */
   if(error != 0) {
      error = 1;
   }
   return error;

}

static void _print_info_label(struct diskinfo *di, int pcount,
                              char **poolnames, int *maxname)
{
   int i, j;
   int tmp;

   *maxname = tmp = 0;

   if(poolnames != NULL) {
      for(i = 0; i < pcount; i++) {
         for(j = 0; j < di->pl_count; j++) {
            if(!strcmp(poolnames[i], di->pl[j].name))
               break;
         }
         if(j != di->pl_count) {
            tmp = strlen(di->pl[j].name);
            if(tmp > *maxname) {
               *maxname = tmp;
            }
         }
      }
   }
   else {
      for(j = 0; j < di->pl_count; j++) {
         tmp = strlen(di->pl[j].name);
         if(tmp > *maxname) {
            *maxname = tmp;
         }
      }
   }

   log_warn("Major Minor Name%*sAlias        "
            "  Capacity  In use   MP Type"
            "   MP Stripe", *maxname > 4 ? (*maxname)-3 : 1, "");

}

static void _print_stats(char *poolname, user_poolstat_t *iostat)
{

   printf("[%10s]", poolname);
   printf("  rd: %d rdblk: %d  wr: %d  wrblk: %d\n", iostat->reads,
         iostat->readblks, iostat->writes, iostat->writeblks);
}

static int _get_capacity(struct user_pool *pl, uint64 *cap, double *hcap,
                         char *ext, int hrd, struct sstring *emsg)
{
   uint64 blocks = 0;
   double tmp = 0, hr = 0;
   int i, j, k;

   for(j = 0; j < pl->num_sps; j++) {
      for(k = 0; k < pl->sp[j].num_devs; k++) {
         blocks += pl->sp[j].devs[k].blocks;
      }
   }
   *cap = blocks;
   *hcap = 0;

   if(hrd) {
      /* FIXME: Assuming 512-byte blocks ATM */
      hr = blocks >> 1;
      tmp = hr;
      for(i = 0; hr > 1 && i <= 8; i++) {
         tmp = hr;
         hr = hr / 1024.0;
      }
      hr = tmp;
      switch(i) {
         case 0:
            *ext = 'b';
            break;
         case 1:
            *ext = 'k';
            break;
         case 2:
            *ext = 'M';
            break;
         case 3:
            *ext = 'G';
            break;
         case 4:
            *ext = 'T';
            break;
         case 5:
            *ext = 'P';
            break;
         case 6:
            *ext = 'E';
            break;
         case 7:
            *ext = 'Z';
            break;
         case '8':
            *ext = 'Y';
            break;
         default:
            *ext = '-';
            break;
      }
      *hcap = hr;
   }
   else {
      *ext = '\0';
   }

   return 0;
}

static int _print_verbose_info(struct user_pool *pl, int hr,
                               struct sstring *emsg)
{
   int error = 0;
   struct user_subpool *sp = NULL;
   struct user_device *dev = NULL;
   uint64 size = 0;
   char *type = NULL;
   char *alias = NULL;
   char ext;
   int j, k, l;
   double hsize;

   get_pool_alias(&alias, pl->minor, emsg);

   log_notice("Pool name             : %s", pl->name);
   log_notice("Pool alias            : %s", alias);
   destroy_pool_alias(alias);

   log_notice("Number of subpools    : %u", pl->num_sps);
   _get_capacity(pl, &size, &hsize, &ext, hr, emsg);
   if(hr != 0) {
      log_notice("Total Capacity        : %0.4g%c", hsize, ext);
   }
   else {
      log_notice("Total Capacity        : %"PRIu64, size);
   }
   log_notice("In use                : %s",
              (pl->in_use) ? "YES" : "NO");
   log_notice("Multipathing Type     : %s",
              (pl->mptype == 0 ? "none" :
               (pl->mptype == 1 ? "failover" : "round-robin")));
   if(pl->mptype > 1) {
      log_notice("Multipathing Stripe   : %d", pl->mpstripe);
   }
   log_notice("Pool ID               : 0x%016"PRIx64, pl->id);
   for (j=0; j < pl->num_sps; j++) {
      sp = &(pl->sp[j]);
      log_notice("Subpool %d", j);
      log_notice("Number of devices     : %u", sp->num_devs);
      log_notice("Stripe size           : %u", sp->striping);
      get_sptype_name(sp->type, &type);
      log_notice("Type                  : %s", type);
      destroy_sptype_name(&type);

      for (k=0; k < sp->num_devs; k++) {
         dev = &(sp->devs[k]);

         log_notice("Device %d", k);
         log_notice("Path count            : %d", dev->path_count);
         log_notice("Good path count       : %d", dev->good_path_count);
         for(l=0; l < dev->path_count; l++){
            char *path = NULL;
            log_notice("Path %d                : %-10s (%s)",
                   l, dev->dev_path[l],
                   (l < dev->good_path_count) ? "GOOD": "BAD ");
            free(path);
         }
         log_notice("Blocks                : %"PRIu64,
                dev->blocks);
      }
   }
   log_notice("");


   return error;
}


static void _print_info(user_pool_t *pl, int humanread, int len)
{
   uint64_t size;
   char *alias,
         ext;
   double hsize;
   struct sstring *emsg = ss_init(STRING_SIZE-1);

   get_pool_alias(&alias, pl->minor, emsg);

   _get_capacity(pl, &size, &hsize, &ext, humanread, emsg);

   if(humanread == 1) {
      log_warn("%-5d %5d %-*s %-12s %9.4g%c%8s %11s %5.0d",
               pl->major, pl->minor, len, pl->name, alias,
               hsize, ext, pl->in_use ? "YES" : "NO",
               (pl->mptype == 0 ? "none" :
                (pl->mptype == 1 ? "failover" : "round-robin")),
               pl->mptype > 1 ? pl->mpstripe : 0);
   }
   else {
      log_warn("%-5d %5d %-*s %-12s %10"PRIu64"%8s %11s %5.0d",
               pl->major, pl->minor, len, pl->name, alias,
               size, pl->in_use ? "YES" : "NO",
               (pl->mptype == 0 ? "none" :
                (pl->mptype == 1 ? "failover" : "round-robin")),
               pl->mptype > 1 ? pl->mpstripe : 0);
   }
   destroy_pool_alias(alias);
}


/*
 * vim: et ts=3 sw=3 noai cin :
 */


