/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**  
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/
/* The ccs_read is meant to provide a CLI to the CCS.  In addition to     **
** providing this functionality, it is also meant to be an implementation **
** guide to others looking to use the CCS.  Reading the included comments **
** in the following code should bring you up to speed quickly on the CCS  **
** API................................................................... **
**                                                                        **
** Additionally, at the end of this file, there is an example CCS file    **
** and more comments on the relationship of information in those files.   */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <asm/errno.h>
#include "copyright.cf"
#include "ccs_release.cf"

/* ccs.h can be found in sistina/ccs/lib.  It contains all the structs    **
** needed for retrieving data in a file managed by the CCS system.  It    **
** also contains the necessary function prototypes....................... */

#include "ccs.h"

void print_usage(FILE *stream);
void print_node(ccs_node_t *cn);

int main(int argc, char *argv[]){
  int  i, error;
  int c;
  char sep='/';
  char *cmd=NULL;
  char *file=NULL;
  char *request=NULL;

  while((c = getopt(argc, argv, "s:hV")) != -1) {
    switch(c){
    case 's':
      if(strlen(optarg) > 1){
	fprintf(stderr,
		"Only a single character is allowed as a separator.\n");
	exit(EXIT_FAILURE);
      }
      sep = optarg[0];
      break;
    case 'h':
      print_usage(stdout);
      exit(EXIT_SUCCESS);
    case 'V':
      printf("%s %s (built %s %s)\n", argv[0], CCS_RELEASE_NAME,
	     __DATE__, __TIME__);
      printf("%s\n", SISTINA_COPYRIGHT);
      exit(EXIT_SUCCESS);
      break;
    default:
      print_usage(stderr);
      exit(EXIT_FAILURE);
    }
  }

  for(i=0; i < (argc-optind); i++){
    if(!i){
      cmd = argv[i+optind];
    } else if((i == 1) && strcmp(cmd, "list")){
      file = argv[i+optind];
    } else if((i == 2) && strcmp(cmd, "file") && strcmp(cmd, "list")){
      request = argv[i+optind];
    } else {
      fprintf(stderr, "Too many arguments.\n");
      fprintf(stderr, "Try 'ccs_read -h' for help.\n");
      exit(EXIT_FAILURE);
    }      
  }

  if((i < 1) ||
     ((i < 2) && strcmp(cmd, "list")) || 
     ((i < 3) && (strcmp(cmd, "file") &&
#ifdef DEBUG
		  strcmp(cmd, "local") &&
#endif
		  strcmp(cmd, "list")
		  ))){
    fprintf(stderr, "Too few arguments.\n");
    fprintf(stderr, "Try 'ccs_read -h' for help.\n");
    exit(EXIT_FAILURE);
  }


  if(!strcmp(cmd, "list")){
    char *data;

    error = get_ccs_file(&data, "list_files");
    if(error < 0){
      fprintf(stderr, "Unable to get list of files: %s\n", strerror(-error));
      exit(EXIT_FAILURE);
    } else {
      printf("%s", data);
      free(data);
    }
  } else if(!strcmp(cmd, "file")){
    char *data;

    /* Getting file data:                                                  **
    **                                                                     **
    ** If you simply want all of the data in the file, it is just a matter **
    ** of calling get_ccs_file.  The first argument is a (char **).  Space **
    ** will be allocated for the data.  The amount of data stored is       **
    ** returned by the function on success.  Otherwise, -1 is returned.    **
    ** The last argument to the function is the name of the file you want  **
    ** retrieved.                                                          **
    **                                                                     **
    ** If there are any errors, any memory that was allocated will be      **
    ** freed before the function returns.  Otherwise, it is the            **
    ** RESPONSIBILITY OF THE CALLER to free the returned data.             */
    error = get_ccs_file(&data, file);
    if(error < 0){
      fprintf(stderr, "Cannot retrieve file \"%s\": %s\n",
	      file, strerror(-error));
      exit(EXIT_FAILURE);
    } else {
      printf("%s", data);
      free(data);
    }
#ifdef DEBUG
  } else if(!strcmp(cmd, "local")){
    char *data;

    /* Getting local file data (data in files on machine... not in CCS):   **
    **                                                                     **
    ** The get_local_file() function works exactly like get_ccs_file()     **
    ** except that it will retrieve a file off of a file system that is    **
    ** available to the machine, rather than from the CCS.  The function   **
    ** is not normally available to user space applications due to the     **
    ** security risks.  Only when compiled with debugging on (make DEBUG=y)**
    ** is this function available.                                         **
    **                                                                     **
    ** Note: It would be exceedingly stupid to call get_local_file() if the**
    ** file being requested is on a non-local file system.                 **
    **                                                                     **
    ** If there are any errors, any memory that was allocated will be      **
    ** freed before the function returns.  Otherwise, it is the            **
    ** RESPONSIBILITY OF THE CALLER to free the returned data.             */
    if(get_local_file(&data, file) < 0){
      fprintf(stderr, "Cannot retrieve file \"%s\"\n", file);
      exit(EXIT_FAILURE);
    } else {
      printf("%s\n", data);
      free(data);
    }
#endif /* DEBUG */
  } else if(!strcmp(cmd, "section")){
    ccs_node_t *root_cn;
    ccs_node_t *cn;
    
    /* Getting a node (a.k.a Section/subsection):                          **
    **                                                                     **
    ** Refer to the section at the end of this file for a description of   **
    ** what nodes are...................................................   */

    /* First, we open the file.  open_ccs_file() takes a double ptr and a  **
    ** filename as arguments.  It returns -EXXX on failure, or 0 on        **
    ** success.  The ccs_node_t double ptr is assigned during the open     **
    ** call and can be passed to other functions, such as find_ccs_int(),  **
    ** to retrieve info                                                    **
    **                                                                     **
    ** The open causes the file to be read in and parsed.  The parsed data **
    ** is held by the CCS system until close_ccs_file() is called.  This   **
    ** reduces network traffic because the ccs system knows how long a     **
    ** particular file is needed.                                          */
    error = open_ccs_file(&root_cn, file);
    
    if(error){
      if(error == -EPARSE){
	fprintf(stderr, "An error occured while parsing the CCS file, %s.\n",
		file);
      } else {
	fprintf(stderr, "Unable to open the CCS file, %s: %s\n",
		file, strerror(-error));
      }
      exit(EXIT_FAILURE);
    }

    /* The next step to retrieving a particular node is to call            **
    ** find_ccs_node().  It's arguments are the root ccs node returned by  **
    ** open_ccs_file(), a request string that is separated by the final    **
    ** argument.                                                           **
    **                                                                     **
    ** Assuming the example config file at the end of this file, if we do  **
    **   find_ccs_node(root_cn, "Sistina/employees", '/')                  **
    ** we will get back the node with a key value of employee and info     **
    ** filled in appropriately.  (Refer to example file for more info.)    */
    cn = find_ccs_node(root_cn, request, sep);

    if(!cn){
      fprintf(stderr, "Cannot find section \"%s\" in file \"%s\"\n",
	      request, file);
      close_ccs_file(root_cn);
      exit(EXIT_FAILURE);
    }

    /* Take a look at this function to see how all the data is pulled out. */
    print_node(cn);

    /* REMEMBER to close the reference to the ccs file.  This will free all**
    ** data kept in cn and do some house keeping on the CCS side.  Do not  **
    ** try to access anything in the root ccs node or call any             **
    ** find_ccs_XXX() functions, as the data will not be valid after close.*/
    close_ccs_file(root_cn);
  } else if(!strcmp(cmd, "string")){
    const char *str;
    ccs_node_t *root_cn;
    ccs_node_t *cn;

    /* again, open the ccs file */
    error = open_ccs_file(&root_cn, file);
    if(error){
      if(error == -EPARSE){
	fprintf(stderr, "An error occured while parsing the CCS file, %s.\n",
		file);
      } else {
	fprintf(stderr, "Unable to open the CCS file, %s: %s\n",
		file, strerror(-error));
      }
      exit(EXIT_FAILURE);
    }

    cn = find_ccs_node(root_cn, request, sep);

    if(!cn){
      fprintf(stderr, "Cannot find the string \"%s\" in file \"%s\"\n",
	      request, file);
      close_ccs_file(root_cn);
      exit(EXIT_FAILURE);
    }
 
    if(cn && cn->v && cn->v->next){
      fprintf(stderr,
	      "The requested value is an array.\n"
	      "Use the 'section' command in place of 'string'.\n");
      close_ccs_file(root_cn);
      exit(EXIT_FAILURE);
    }
    /* Make your query...   The last argument to these functions is a **
    ** default value to return in case of failure.  Here I return a   **
    ** value even if nothing is found, but you can use NULL, or       **
    ** whatever...................................................... */
    if(!cn->v || (cn->v && cn->v->type != CCS_STRING)){
      fprintf(stderr, "\"%s\" is not of type string.\n", request);
      close_ccs_file(root_cn);
      exit(EXIT_FAILURE);
    }
    str = find_ccs_str(root_cn, request, sep, "<No such string>");
    if(!strcmp(str, "<No such string>")){
      fprintf(stderr, "Cannot find the string \"%s\" in the file \"%s\"\n",
	      request, file);
      close_ccs_file(root_cn);
      exit(EXIT_FAILURE);
    }
    printf("%s\n", str);
    
    /* close the ccs file */
    close_ccs_file(root_cn);
  } else if(!strcmp(cmd, "int")){
    ccs_node_t *root_cn;
    ccs_node_t *cn;

    /* again, open the ccs file */
    error = open_ccs_file(&root_cn, file);

    if(error){
      if(error == -EPARSE){
	fprintf(stderr, "An error occured while parsing the CCS file, %s.\n",
		file);
      } else {
	fprintf(stderr, "Unable to open the CCS file, %s: %s\n",
		file, strerror(-error));
      }
      exit(EXIT_FAILURE);
    }

    cn = find_ccs_node(root_cn, request, sep);
    if(!cn){
      fprintf(stderr, "Cannot find the int \"%s\" in the file \"%s\"\n",
	      request, file);
      close_ccs_file(root_cn);
      exit(EXIT_FAILURE);      
    } else if(cn->v && cn->v->next){
      fprintf(stderr,
	      "The requested value is an array.\n"
	      "Use the 'section' command in place of 'string'.\n");
      close_ccs_file(root_cn);
      exit(EXIT_FAILURE);
    }

    /* Make your query... */
    if(!cn->v || (cn->v && cn->v->type != CCS_INT)){
      fprintf(stderr, "\"%s\" is not of type int.\n", request);
      close_ccs_file(root_cn);
      exit(EXIT_FAILURE);
    }
    printf("%d\n", find_ccs_int(root_cn, request, sep, -1));
    
    /* close the ccs file */
    close_ccs_file(root_cn);
  } else if(!strcmp(cmd, "boolean")){
    int rtn;
    ccs_node_t *root_cn;
    ccs_node_t *cn;

    /* again, open the ccs file */
    error = open_ccs_file(&root_cn, file);

    if(error){
      if(error == -EPARSE){
	fprintf(stderr, "An error occured while parsing the CCS file, %s.\n",
		file);
      } else {
	fprintf(stderr, "Unable to open the CCS file, %s: %s\n",
		file, strerror(-error));
      }
      exit(EXIT_FAILURE);
    }

    cn = find_ccs_node(root_cn, request, sep);
    if(!cn){
      fprintf(stderr, "Cannot find the boolean \"%s\" in the file \"%s\"\n",
	      request, file);
      close_ccs_file(root_cn);
      exit(EXIT_FAILURE);      
    } else if(cn->v && cn->v->next){
      fprintf(stderr,
	      "The requested value is an array.\n"
	      "Use the 'section' command in place of 'string'.\n");
      close_ccs_file(root_cn);
      exit(EXIT_FAILURE);
    }

    rtn = find_ccs_bool(root_cn, request, sep, -1);
    if(rtn == -1){
      fprintf(stderr, "\"%s\" is not a valid boolean value.\n", request);
      close_ccs_file(root_cn);
      exit(EXIT_FAILURE);
    }
    /* Make your query... */
    printf("%s\n", rtn ? "TRUE": "FALSE");
    
    /* close the ccs file */
    close_ccs_file(root_cn);
  } else {
    fprintf(stderr, "Unrecognized option.\n\n");
    print_usage(stderr);
    exit(EXIT_FAILURE);
  }
  exit(EXIT_SUCCESS);
}


void print_usage(FILE *stream){
  fprintf(stream, "Print cluster configuration file:\n"
	  "  ccs_read [option] <command> <filename> [request]\n\n"
	  "  options:\n"
	  "\t-h        Help.\n"
	  "\t-s <sep>  Specify the separator in the request string.\n"
	  "\t-V        Print version information.\n"
	  "\n"
	  "  commands:\n"
	  "\tfile      Print out a file managed by the CCS.\n"
	  "\tlist      Print a list of files managed by the CCS.\n"
	  "\tsection   Print the section found in a CCS file.\n"
	  "\tstring    Print out a key identified string.\n"
	  "\tint       Print out a key identified integer.\n"
	  "\tboolean   Print out a key identified boolean.\n"
	  "\n"
	  "  filename:\n"
	  "\tA required argument by all commands except \"list\", which\n"
	  "\tspecifies the CCS file to access.\n"
	  "\n"
	  "  request:\n"
	  "\tA required argument by all commands except \"list\" and \n"
	  "\t\"file\", which specifies what to extract from the CCS file.\n"
	  );
}

void _print_node(ccs_node_t *cn, int level){
  int i;

  for(i=level; i; i--) printf("  ");
  printf("%s", cn->key);
  if(cn->child){
    printf(" {\n");
    _print_node(cn->child, level+1);
    for(i=level; i; i--) printf("  ");
    printf("}\n");
  } else {
    ccs_value_t *val;
    int array;
    
    if(cn->v){
      array = (cn->v->next)?1:0;
      printf("%s", (array)? " = [ ": " = ");
      
      for(val = cn->v; val; val = val->next){
	if(val->type == CCS_STRING){
	  printf("\"%s\"%s", val->v.str, (array && val->next)?", ":"");
	} else if(val->type == CCS_FLOAT){
	  printf("%f", val->v.r);
	  printf("%s", (array && val->next)?", ":"");	
	} else if(val->type == CCS_INT){
	  printf("%d%s", val->v.i, (array && val->next)?", ":"");
	} else {
	  fprintf(stderr, "\nUnknown value type.\n");
	}
      }
      printf("%s", array ? " ]\n":"\n");
    } else {
      printf(" { }\n");
    }
  }

  if(level && cn->sib){
    _print_node(cn->sib, level);
  }
}


void print_node(ccs_node_t *cn){
  _print_node(cn, 0);
}
