/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "servd.h"
#include "trans.h"
#include "device.h"
#include "gserv.h"
#include "fence.h"
#include "fixed_div64.h"
#include <linux/utsname.h>

#define BH_GnbdInUse BH_PrivateStart
#define buffer_gnbd_in_use(bh) __buffer_state((bh),GnbdInUse)

static void
gnbd_release_buffer(struct buffer_head *bh)
{
	clear_bit(BH_GnbdInUse, &bh->b_state);
	smp_mb__after_clear_bit();
	if (waitqueue_active(&bh->b_wait))
		wake_up(&bh->b_wait);
}

static __inline__ void
gnbd_use_buffer(struct buffer_head *bh)
{
	while (test_and_set_bit(BH_GnbdInUse, &bh->b_state))
		wait_event(bh->b_wait, (!buffer_gnbd_in_use(bh)));
}

LIST_HEAD(gserv_list);
struct semaphore gserv_list_lk;	/* always lock after device lock, if necessary */
wait_queue_head_t gserv_sleep;
atomic_t gserv_count;

void
init_gserv(void)
{
	init_MUTEX(&gserv_list_lk);
	init_waitqueue_head(&gserv_sleep);
	atomic_set(&gserv_count, 0);
}

int
add_gserv_info(struct socket *sock, u32 client_ip, char *name,
	       gserv_info_t ** gserv_info)
{
	int err;
	gserv_info_t *info;
	dev_info_t *dev;
	char ip_str[16];

	*gserv_info = NULL;
	err = get_peer_ip(sock, &client_ip);
	if (err)
		goto out;

	dev = find_and_lock_device(name);
	if (dev == NULL) {
		printe("client %s requesting device %s that is not exported\n",
		       beip_to_str(client_ip, ip_str), name);
		err = -ENODEV;
		goto out;
	}

	info = kmalloc(sizeof (gserv_info_t), GFP_KERNEL);
	if (!info) {
		printe("couldn't allocate memory for server info\n");
		err = -ENOMEM;
		goto out_lock;
	}
	info->client_ip = client_ip;
	info->dev = dev;
	info->sock = sock;
	init_MUTEX(&info->sender_lk);
	info->task = current;
	info->pid = (long) (current->pid);

	down(&gserv_list_lk);
	list_add(&info->list, &gserv_list);
	up(&gserv_list_lk);

	*gserv_info = info;

      out_lock:
	up(&info->dev->lock);

      out:
	return err;
}

void
remove_gserv_info(gserv_info_t * gserv_info)
{

	down(&gserv_list_lk);
	list_del(&gserv_info->list);
	up(&gserv_list_lk);
	kfree(gserv_info);

	return;
}

void
validate_gserv(void)
{
	int err = 0;
	list_t *list_item;
	gserv_info_t *gserv_info;
	trans_hdr_t ping;
	char ip_str[16];

	memset(&ping, 0, sizeof (ping));
	ping.type = gnbd_ping;

	down(&gserv_list_lk);
	list_for_each(list_item, &gserv_list) {
		gserv_info = list_entry(list_item, gserv_info_t, list);
		down(&gserv_info->sender_lk);
		err =
		    transfer_info(SEND, gserv_info->sock, &ping, sizeof (ping));
		if (err) {
			printe("ping to %s failed at line %d (%d)\n",
			       beip_to_str(gserv_info->client_ip, ip_str),
			       __LINE__, err);
			err = send_sig(SIGINT, gserv_info->task, 0);
			if (err)
				printe
				    ("error sending SIGINT to gserv process pid %ld: (%d)\n",
				     gserv_info->pid, err);
		}
		up(&gserv_info->sender_lk);
	}
	up(&gserv_list_lk);
	return;
}

int
find_gserv_info(u32 client_ip, dev_info_t * dev)
{
	int found = 0;
	list_t *list_item;
	gserv_info_t *gserv_info = NULL;

	down(&gserv_list_lk);
	list_for_each(list_item, &gserv_list) {
		gserv_info = list_entry(list_item, gserv_info_t, list);
		if ((!client_ip || client_ip == gserv_info->client_ip) &&
		    (!dev || dev == gserv_info->dev)) {
			found = 1;
			break;
		}
	}
	up(&gserv_list_lk);
	return found;
}

int
kill_gserv(u32 client_ip, dev_info_t * dev)
{
	int err = 0;
	list_t *list_item;
	gserv_info_t *gserv_info = NULL;

	down(&gserv_list_lk);
	list_for_each(list_item, &gserv_list) {
		gserv_info = list_entry(list_item, gserv_info_t, list);
		if ((!client_ip || client_ip == gserv_info->client_ip) &&
		    (!dev || dev == gserv_info->dev)) {
			err = send_sig(SIGINT, gserv_info->task, 0);
			if (err) {
				printe
				    ("error sending SIGINT to gserv process, pid %ld. (%d)\n",
				     gserv_info->pid, err);
				up(&gserv_list_lk);
				goto out;
			}
		}
	}
	up(&gserv_list_lk);

	while (find_gserv_info(client_ip, dev)) {
		DECLARE_WAITQUEUE(__wait_chan, current);
		current->state = TASK_UNINTERRUPTIBLE;
		add_wait_queue(&gserv_sleep, &__wait_chan);
		if (find_gserv_info(client_ip, dev))
			schedule();
		remove_wait_queue(&gserv_sleep, &__wait_chan);
		current->state = TASK_RUNNING;
	}

      out:
	return err;
}

void
kill_bad_gserv(u32 client_ip, char *name)
{
	dev_info_t *dev;
	char ip_str[16];

	dev = find_and_lock_device(name);
	if (dev == NULL) {
		printe
		    ("client %s can't kill gserv of a device %s that is not exported\n",
		     beip_to_str(client_ip, ip_str), name);
		return;
	}
	else
		while (kill_gserv(client_ip, dev)) ;
	up(&dev->lock);
	return;
}

int
gserv_login(struct socket *sock, u32 client_ip, gserv_info_t ** gserv_info)
{
	int err;
	login_req_t login_request;
	login_reply_t login_reply;
	gserv_info_t *info;
	char ip_str[16];

	*gserv_info = NULL;
	err =
	    timed_transfer_info(RECV, sock, &login_request,
				sizeof (login_request));
	if (err) {
		printe("transfer of login request from %s failed (%d)\n",
		       beip_to_str(client_ip, ip_str), err);
		goto exit;
	}
	login_request.timestamp = be64_to_cpu(login_request.timestamp);
	login_request.version = be64_to_cpu(login_request.version);

	err = update_timestamp_list(client_ip, login_request.timestamp);
	if (err)
		goto fail_reply;

	err = add_gserv_info(sock, client_ip, login_request.devname, &info);
	if (err)
		goto fail_reply;

	if (check_banned_list(client_ip)) {
		printe("client %s is banned. Canceling login\n",
		       beip_to_str(client_ip, ip_str));
		remove_gserv_info(info);
		err = -EPERM;
		goto fail_reply;
	}

	login_reply.err = 0;
	login_reply.devsize = cpu_to_be64(info->dev->size);
	login_reply.timeout = cpu_to_be32(info->dev->timeout);
	login_reply.name_size = cpu_to_be16(buflen(system_utsname.nodename));
	login_reply.version = cpu_to_be16(PROTOCOL_VERSION);
	login_reply.readonly = info->dev->readonly;

	down(&info->sender_lk);
	err = transfer_info(SEND, sock, &login_reply, sizeof (login_reply));
	up(&info->sender_lk);
	if (err) {
		printe("transfer of login reply to %s failed (%d)\n",
		       beip_to_str(client_ip, ip_str), err);
		remove_gserv_info(info);
		goto exit;
	}

	down(&info->sender_lk);
	err = transfer_info(SEND, sock, system_utsname.nodename,
			    buflen(system_utsname.nodename));
	up(&info->sender_lk);
	if (err) {
		printe("transfer of node name to %s failed (%d)\n",
		       beip_to_str(client_ip, ip_str), err);
		remove_gserv_info(info);
		goto exit;
	}
	*gserv_info = info;

      exit:
	return err;

      fail_reply:
	login_reply.err = -err;
	transfer_info(SEND, sock, &login_reply, sizeof (login_reply));
	return err;
}

#define LOCKED_SEND(action) \
do{ \
  err = (action); \
  if (err){ \
    up(&gserv_info->sender_lk); \
    printe("while serving device %s,\n" \
          "transfer with %s failed at line %d (%d)\n", gserv_info->dev->name, \
          beip_to_str(gserv_info->client_ip, ip_str), __LINE__, err); \
    goto exit_mem; \
  } \
} while(0)

#define DO_SEND(action) \
do{ \
  down(&gserv_info->sender_lk); \
  err = (action); \
  up(&gserv_info->sender_lk); \
  if (err){ \
    printe("while serving device %s,\n" \
          "transfer with %s failed at line %d (%d)\n", gserv_info->dev->name, \
          beip_to_str(gserv_info->client_ip, ip_str), __LINE__, err); \
    goto exit_mem; \
  } \
} while(0)

#define DO_RECV(action) \
do{ \
  err = (action); \
  if (err){ \
    printe("while serving device %s,\n" \
          "transfer with %s failed at line %d (%d)\n", gserv_info->dev->name, \
          beip_to_str(gserv_info->client_ip, ip_str), __LINE__, err); \
    goto exit_mem; \
  } \
} while(0)

#define FIRST_BH 0
#define LAST_BH 1

int
uncached_gserv(gserv_info_t * gserv_info)
{
	int err;
	char ip_str[16];
	struct buffer_head **bh;
	int bh_index;
	trans_hdr_t request;
	signal_info_t sig_info;
	int blksize;
	u64 block_64;
	u64 req_start;
	u32 req_len;
	int trans_len;
	int start_offset;
	int end_block;
	int end_len;
	int blocks_left;
	int block;
	struct buffer_head *tmp_bh[2];
	int i;
	struct socket *sock;

	sock = gserv_info->sock;
	blksize = gserv_info->dev->blksize;
	bh = kmalloc(128 * sizeof (struct buffer_head *), GFP_KERNEL);
	if (!bh) {
		printe("couldn't allocate memory for buffer heads\n");
		err = -ENOMEM;
		goto exit;
	}

	while (1) {
		DO_RECV(transfer_info(RECV, sock, &request, sizeof (request)));
		req_start = be64_to_cpu(request.offset);
		req_len = be32_to_cpu(request.len);
		block_64 = req_start;
		start_offset = (int) do_div(block_64, blksize);
		block = (int) block_64;
		block_64 = req_start + req_len - 1;
		end_len = (int) do_div(block_64, blksize);
		end_block = (int) block_64;
		end_len += 1;
		blocks_left = end_block - block + 1;

		if ((request.type == gnbd_read_req
		     || request.type == gnbd_write_req)
		    && req_start + req_len > gserv_info->dev->size) {
			printe("request from %s past the end of the device\n",
			       beip_to_str(gserv_info->client_ip, ip_str));
			request.type = gnbd_error;
			request.err = error_invalid_request;
			DO_SEND(transfer_info
				(SEND, sock, &request, sizeof (request)));
			goto exit_mem;
		}
		switch (request.type) {
		case gnbd_read_req:
			request.type = gnbd_read_reply;
			down(&gserv_info->sender_lk);
			LOCKED_SEND(transfer_info
				    (SEND, sock, &request, sizeof (request)));
			while (blocks_left) {
				for (bh_index = 0;
				     blocks_left && bh_index < 128;
				     blocks_left--, bh_index++, block++) {
					bh[bh_index] =
					    getblk(gserv_info->dev->rdev, block,
						   blksize);
					gnbd_use_buffer(bh[bh_index]);
					mark_buffer_uptodate(bh[bh_index], 0);
				}
				ll_rw_block(READ, bh_index, &bh[0]);
				for (i = 0; i < bh_index; i++) {
					wait_on_buffer(bh[i]);
					if (!buffer_uptodate(bh[i])) {
						printe
						    ("could not get buffer %lu from the device\n",
						     (unsigned long) bh[i]->
						     b_data);
						/* It would be nice to be able to notify the client, but it
						   is impossible, so I'll just shut down */
						err = -EIO;
						goto read_release;
					}
				}
				/* since the client can request up to 128 * it's blksize bytes of 
				   data, and the server can send up to 128 * it's blksize bytes of
				   data, this is correct, even if it looks scary.
				   If the 128 gets changed, it must be changed in both places. */
				if (blocks_left)
					trans_len = bh_index * blksize;
				else
					trans_len = req_len;
				err =
				    trans_bh_data(SEND, sock, bh, bh_index,
						  start_offset, end_len,
						  trans_len);
				if (err)
					printe("while serving device %s,\n"
					       "transfer with %s failed at line %d (%d)\n",
					       gserv_info->dev->name,
					       beip_to_str(gserv_info->
							   client_ip, ip_str),
					       __LINE__, err);

			      read_release:
				for (i = 0; i < bh_index; i++){
					gnbd_release_buffer(bh[i]);
					brelse(bh[i]);
				}
				if (err) {
					up(&gserv_info->sender_lk);
					goto exit_mem;
				}
			}
			up(&gserv_info->sender_lk);
			break;
		case gnbd_write_req:
			if (gserv_info->dev->readonly) {
				printe("write request to a read only device\n");
				request.type = gnbd_error;
				request.err = error_invalid_request;
				DO_SEND(transfer_info
					(SEND, sock, &request,
					 sizeof (request)));
				goto exit_mem;
			}
			i = 0;
			/* Read in the first and last blocks if the request doesn't align
			   with block boundaries. This sorta seems excessive for the
			   time savings */
			tmp_bh[FIRST_BH] =
			    getblk(gserv_info->dev->rdev, block, blksize);
			bh_index = 1;
			if (start_offset > 0) {
				mark_buffer_uptodate(tmp_bh[FIRST_BH], 0);
				bh_index = 0;
				i++;
			}
			if (block != end_block) {
				tmp_bh[LAST_BH] =
				    getblk(gserv_info->dev->rdev, end_block,
					   blksize);
				if (end_len < blksize) {
					mark_buffer_uptodate(tmp_bh[LAST_BH],
							     0);
					i++;
				}
			}
			if (i)
				ll_rw_block(READ, i, &tmp_bh[bh_index]);

			/* special case for first block */
			bh[0] = tmp_bh[FIRST_BH];
			block++;
			blocks_left--;
			bh_index = 1;

			do {
				while (blocks_left > 1 && bh_index < 128) {
					bh[bh_index++] =
					    getblk(gserv_info->dev->rdev,
						   block++, blksize);
					blocks_left--;
				}
				/* special case for the last block */
				if (blocks_left == 1 && bh_index < 128) {
					bh[bh_index++] = tmp_bh[LAST_BH];
					blocks_left--;
				}
				if (blocks_left)
					trans_len = bh_index * blksize;
				else
					trans_len = req_len;
				err =
				    trans_bh_data(RECV, sock, bh, bh_index,
						  start_offset, end_len,
						  trans_len);
				if (err) {
					printe("while serving device %s,\n"
					       "transfer with %s failed at line %d (%d)\n",
					       gserv_info->dev->name,
					       beip_to_str(gserv_info->
							   client_ip, ip_str),
					       __LINE__, err);
					goto write_release;
				}
				for (i = 0; i < bh_index; i++) {
					mark_buffer_uptodate(bh[i], 1);
					mark_buffer_dirty(bh[i]);
				}
				ll_rw_block(WRITE, bh_index, bh);
				for (i = 0; i < bh_index; i++) {
					wait_on_buffer(bh[i]);
					if (!buffer_uptodate(bh[i])) {
						printe
						    ("could not write buffer %lu to the device\n",
						     (unsigned long) (bh[i]->
								      b_data));
						err = -EIO;
						request.type = gnbd_error;
						request.err =
						    error_server_failure;
						DO_SEND(transfer_info
							(SEND, sock, &request,
							 sizeof (request)));
						goto write_release;
					}
					brelse(bh[i]);
				}
				bh_index = 0;
			} while (blocks_left);
			request.type = gnbd_write_reply;
			DO_SEND(transfer_info
				(SEND, sock, &request, sizeof (request)));
			break;
		case gnbd_ping:
			DO_SEND(transfer_info
				(SEND, sock, &request, sizeof (request)));
			break;
		case gnbd_shutdown:
			goto exit_mem;
		default:
			printe
			    ("Unknown request type (%d) from %s, shutting down\n",
			     (int) request.type,
			     beip_to_str(gserv_info->client_ip, ip_str));
			goto exit_mem;
		}
	}

      write_release:
	for (; i < bh_index; i++)
		brelse(bh[i]);

      exit_mem:
	kfree(bh);
      exit:
	block_signals(&sig_info);
	request.type = gnbd_shutdown;
	down(&gserv_info->sender_lk);
	transfer_info(SEND, sock, &request, sizeof (trans_hdr_t));
	up(&gserv_info->sender_lk);

	return err;
}

int
cached_gserv(gserv_info_t * gserv_info)
{
	int err;
	char ip_str[16];
	char *buffer = NULL;
	trans_hdr_t request;
	signal_info_t sig_info;
	int bufsize;
	loff_t file_offset;
	size_t req_len;
	u32 trans_len;
	struct file_operations *ops;
	char *ptr;
	mm_segment_t segment_sav;
	struct socket *sock;

	sock = gserv_info->sock;
	ops = gserv_info->dev->file->f_op;
	bufsize = 512 * 128;	/* for 128 blocks of 512 bytes */
	buffer = (char *) kmalloc(bufsize, GFP_KERNEL);
	if (!buffer) {
		printe("couldn't allocate memory for the request buffer\n");
		err = -ENOMEM;
		goto exit;
	}

	while (1) {
		DO_RECV(transfer_info(RECV, sock, &request, sizeof (request)));
		req_len = (size_t) be32_to_cpu(request.len);
		file_offset = (loff_t) be64_to_cpu(request.offset);

		if ((request.type == gnbd_read_req
		     || request.type == gnbd_write_req)
		    && (u64) file_offset + req_len > gserv_info->dev->size) {
			printe("request from %s past the end of the device\n",
			       beip_to_str(gserv_info->client_ip, ip_str));
			request.type = gnbd_error;
			request.err = error_invalid_request;
			DO_SEND(transfer_info
				(SEND, sock, &request, sizeof (request)));
			goto exit_mem;
		}

		switch (request.type) {
		case gnbd_read_req:
			request.type = gnbd_read_reply;
			down(&gserv_info->sender_lk);
			LOCKED_SEND(transfer_info
				    (SEND, sock, &request, sizeof (request)));
			while (req_len > 0) {
				int tmp_len;

				trans_len =
				    (req_len > bufsize) ? bufsize : req_len;
				ptr = buffer;
				req_len -= trans_len;
				tmp_len = trans_len;
				while (tmp_len > 0) {
					int result;

					segment_sav = get_fs();
					set_fs(get_ds());
					result =
					    ops->read(gserv_info->dev->file,
						      ptr, tmp_len,
						      &file_offset);
					set_fs(get_ds());
					if (result < 0) {
						up(&gserv_info->sender_lk);
						printe
						    ("error in file read (%d)\n",
						     result);
						/* I can't warn the client, so I'll just exit */
						goto exit_mem;
					}

					tmp_len -= result;
					ptr += result;
				}
				LOCKED_SEND(transfer_info
					    (SEND, sock, buffer, trans_len));
			}
			up(&gserv_info->sender_lk);
			break;
		case gnbd_write_req:
			if (gserv_info->dev->readonly) {
				printe("write request to a read only device\n");
				request.type = gnbd_error;
				request.err = error_invalid_request;
				DO_SEND(transfer_info
					(SEND, sock, &request,
					 sizeof (request)));
				goto exit_mem;
			}
			while (req_len > 0) {
				trans_len =
				    (req_len > bufsize) ? bufsize : req_len;
				ptr = buffer;
				req_len -= trans_len;
				DO_RECV(transfer_info
					(RECV, sock, ptr, trans_len));
				while (trans_len > 0) {
					int result;

					segment_sav = get_fs();
					set_fs(get_ds());
					result =
					    ops->write(gserv_info->dev->file,
						       ptr, trans_len,
						       &file_offset);
					set_fs(get_ds());
					if (result < 0) {
						printe
						    ("error in file write (%d)\n",
						     result);
						request.type = gnbd_error;
						request.err =
						    error_server_failure;
						DO_SEND(transfer_info
							(SEND, sock, &request,
							 sizeof (request)));
						goto exit_mem;
					}

					trans_len -= result;
					ptr += result;
				}
			}
			request.type = gnbd_write_reply;
			DO_SEND(transfer_info
				(SEND, sock, &request, sizeof (request)));
			break;
		case gnbd_ping:
			DO_SEND(transfer_info
				(SEND, sock, &request, sizeof (request)));
			break;
		case gnbd_shutdown:
			goto exit_mem;
		default:
			printe
			    ("Unknown request type (%d) from %s, shutting down\n",
			     (int) request.type,
			     beip_to_str(gserv_info->client_ip, ip_str));
			goto exit_mem;
		}
	}

      exit_mem:
	kfree(buffer);
      exit:
	block_signals(&sig_info);
	request.type = gnbd_shutdown;
	down(&gserv_info->sender_lk);
	transfer_info(SEND, sock, &request, sizeof (trans_hdr_t));
	up(&gserv_info->sender_lk);

	return err;
}

int
gserv(void *args)
{
	int err;
	struct socket *sock = (struct socket *) args;
	gserv_info_t *gserv_info;
	signal_info_t sig_info;
	u32 client_ip;

	gnbd_daemonize("gserv");
	reparent_to_init();
	err = get_peer_ip(sock, &client_ip);
	if (err)
		goto exit;
	sprintf(current->comm, "gserv%08lx",
		(unsigned long) be32_to_cpu(client_ip));
	setup_blocking(&sig_info);
	err = gserv_login(sock, client_ip, &gserv_info);
	if (err)
		goto exit;

	if (cached(gserv_info->dev))
		err = cached_gserv(gserv_info);
	else
		err = uncached_gserv(gserv_info);

	remove_gserv_info(gserv_info);

      exit:
	sock_release(sock);
	DEC_GSERV_COUNT;
	wake_up(&gserv_sleep);

	return err;
}
