/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "gnbd.h"
#include "trans.h"
#include "macros.h"
#include "libgulm.h"
#include "monitor.h"
#include <linux/gnbd_request.h>

gulm_interface_p monitor_interface;
struct completion monitor_wait;
struct semaphore monitor_lk;
int monitor_status;

void
init_monitor(void)
{
	init_MUTEX(&monitor_lk);
	monitor_status = MONITOR_STOPPED;
}

int
monitor_login_reply(void *misc, u64 gen, u32 err, u32 rank, u8 state)
{
	if (err) {
		printe("Core login returned error (%d)\n", err);
		monitor_status = MONITOR_SHUTTING_DOWN;
	} else
		printd("logged into local core\n");

	return err;
}

int
monitor_logout_reply(void *misc)
{
	monitor_status = MONITOR_SHUTTING_DOWN;
	printd("logged out of local core\n");
	return 0;
}

int
monitor_nodechange(void *misc, char *nodename,
		   uint32_t nodeip, uint8_t nodestate)
{
	if (nodestate == lg_core_Fenced) {
		down_read(&gnbd_devices_lk);
		notify_devices(nodeip);
		up_read(&gnbd_devices_lk);
	}
	return 0;
}

int
monitor_error(void *misc, u32 err)
{
	printe("got error code %d from gulm callback\n", err);
	return 0;
}

static lg_core_callbacks_t monitor_cb = {
	login_reply:monitor_login_reply,
	logout_reply:monitor_logout_reply,
	nodechange:monitor_nodechange,
	error:monitor_error
};

int
monitor_msg_loop(void)
{
	int err = 0;

	while (monitor_status == MONITOR_RUNNING) {
		err =
		    lg_core_handle_messages(monitor_interface, &monitor_cb,
					    NULL);
		if (err) {
			printe("error handling gulm messages %d\n", err);
			break;
		}
	}

	return err;
}

int
monitor(void *unused)
{
	int err;

	gnbd_daemonize("gnbd_monitor");
	complete(&monitor_wait);

	err = monitor_msg_loop();

	monitor_status = MONITOR_STOPPED;
	complete(&monitor_wait);

	return err;
}

int
start_monitor(gnbd_t * dev)
{
	int err = 0;

	if (gnbd_cached(dev) || dev->readonly)
		goto exit;
	down(&monitor_lk);
	if (monitor_status != MONITOR_STOPPED) {
		printd("gnbd_monitor is already running\n");
		goto fail;
	}
	init_completion(&monitor_wait);
	err = lg_initialize(&monitor_interface, NULL, "GNBD Monitor Interface");
	if (err) {
		printe("couldn't initialize the interface (%d)\n", err);
		goto fail;
	}
	err = lg_core_login(monitor_interface, FALSE);
	if (err) {
		printe("couldn't login to lock_gulmd (%d). Is it running?\n",
		       err);
		goto interface_fail;
	}
	monitor_status = MONITOR_RUNNING;
	err = kernel_thread(monitor, NULL, 0);
	if (err < 0) {
		printe("couldn't start the gnbd_monitor process (%d)\n", err);
		goto logged_in_fail;
	}
	wait_for_completion(&monitor_wait);

	up(&monitor_lk);
      exit:
	return 0;

      logged_in_fail:
	lg_core_logout(monitor_interface);
	monitor_msg_loop();
      interface_fail:
	lg_release(monitor_interface);
      fail:
	up(&monitor_lk);
	return err;
}

int
needs_monitor(void)
{
	int needs = FALSE;
	int i;

	for (i = 0; i < GNBD_MAX; i++) {
		if (!gnbd_devices[i])
			continue;
		if (gnbd_connected(gnbd_devices[i]) != INFO_DISCONNECTED &&
		    !gnbd_cached(gnbd_devices[i])
		    && !gnbd_devices[i]->readonly) {
			needs = TRUE;
			break;
		}
	}

	return needs;
}

void
stop_monitor(void)
{
	down(&monitor_lk);
	if (!needs_monitor() && monitor_status == MONITOR_RUNNING) {
		lg_core_logout(monitor_interface);
		wait_for_completion(&monitor_wait);
		lg_release(monitor_interface);
	}
	up(&monitor_lk);
}

#define DO_TRANS(action) \
do { \
  char __ip_str[16]; \
  err = (action); \
  if (err){ \
    printe("kill gserv request to %s failed at line %d (%d)\n", \
           beip_to_str(dev->servip, __ip_str), __LINE__, err); \
    goto must_expire; \
  } \
} while (0)

int
send_expire_request(gnbd_t * dev)
{
	int i, err = 0;
	struct socket *sock;
	device_req_t kill_req;
	u32 msg;

	down_read(&gnbd_devices_lk);
	down(&monitor_lk);
	msg = GNBD_SERV_KILL_GSERV;
	strncpy(kill_req.name, gnbd_device_names[gnbd_get_minor(dev)], 255);
	kill_req.name[255] = 0;
	for (i = 1; i < GNBD_MAX; i++) {
		if (gnbd_devices[i] == NULL || !gnbd_devices[i]->timeout ||
		    gnbd_connected(gnbd_devices[i]) != INFO_CONNECTED ||
		    gnbd_devices[i]->readonly)
			continue;
		err = gnbd_send_ping(gnbd_devices[i]);
		if (!err)
			break;
	}
	if (i >= GNBD_MAX) {
		printe
		    ("No other servers to failover to. Aborting expire of %s\n",
		     kill_req.name);
		err = EXPIRE_FAILED;
		goto exit;
	}
	if (timed_connect_to_server(dev->servip, dev->servport, &sock) == 0) {
		DO_TRANS(transfer_be32(SEND, sock, &msg));
		DO_TRANS(transfer_info
			 (SEND, sock, &kill_req, sizeof (kill_req)));
		DO_TRANS(timed_transfer_be32(RECV, sock, &msg));
		printe("%s is still active, aborting expire\n", kill_req.name);
		err = DONT_WAIT;
		goto exit;
	}
      must_expire:
	printe("force expiring %s\n", dev->server_name);
	err = lg_core_forceexpire(monitor_interface, dev->server_name);
	if (err) {
		printe("can't send force expire to core (%d)\n", err);
		err = EXPIRE_FAILED;
	}

      exit:
	up(&monitor_lk);
	up_read(&gnbd_devices_lk);
	return err;
}
