/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

/*
* This is where all of the actual data io between server and client
* happens.
*
*/
#include <asm/uaccess.h>
#include <linux/smp_lock.h>
#include "trans.h"

#include <linux/in.h>
#include <net/sock.h>

#include "macros.h"

int do_transfer(struct socket *sock, struct iovec *iov, int n, int size,
		int dir, unsigned long timeout);

int
get_peer_ip(struct socket *sock, u32 * ip)
{
	struct sockaddr_in peeraddr;
	int len;
	int err;

	err = sock->ops->getname(sock, (struct sockaddr *) &peeraddr, &len, 1);
	if (err) {
		*ip = 0;
		printe("can't get peer ip from socket\n");
	} else
		*ip = peeraddr.sin_addr.s_addr;

	return err;
}

/******************************************************************************
 * ip_to_str
 * 
 * takes a cpu byte-ordered u32 and a buffer, and fills the buffer with
 * an ip address.
 */
char *
ip_to_str(uint32_t ip, char *buf)
{
	int i;
	char *p;

	p = buf;

	for (i = 3; i >= 0; i--) {
		p += sprintf(p, "%d", (ip >> (8 * i)) & 0xFF);
		if (i > 0)
			*(p++) = '.';
	}
	return buf;
}

static void
send_sigusr1(unsigned long arg)
{
	struct task_struct *tsk = (struct task_struct *) arg;
	send_sig(SIGUSR1, tsk, 0);
}

/******************************************************************************
 * connect_to_server()
 * ip in big endian, port in cpu format
 * only set the timeout not equal to zero when you own the process that is
 * creating the socket.  Otherwise, you might mess with the process' signal
 * handling.
 */
int
__connect_to_gnbd_server(u32 ip, u16 port, struct socket **connection,
			 int timeout)
{
	struct sockaddr_in addr;
	struct socket *sock;
	struct timer_list timer;
	sigset_t blocked_sav;
	unsigned long flags;
	char ip_str[16];
	int err;

	printd("connecting to server at %s:%d\n", beip_to_str(ip, ip_str),
	       port);
	*connection = NULL;
	err = sock_create(AF_INET, SOCK_STREAM, 0, &sock);
	if (err) {
		printe("failed to create socket (%d)\n", err);
		goto fail;
	}
	sock->sk->allocation = GFP_NOIO;
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = ip;
	addr.sin_port = htons(port);

	if (timeout) {
		init_timer(&timer);
		timer.function = send_sigusr1;
		timer.data = (unsigned long) current;
	}
	spin_lock_irqsave(&current->sighand->siglock, flags);
	blocked_sav = current->blocked;
	sigdelsetmask(&current->blocked, sigmask(SIGUSR1));
	recalc_sigpending();
	spin_unlock_irqrestore(&current->sighand->siglock, flags);

	if (timeout)
		mod_timer(&timer, jiffies + (timeout * HZ));
	err =
	    sock->ops->connect(sock, (struct sockaddr *) &addr, sizeof (addr),
			       0);
	if (timeout)
		del_timer(&timer);

	spin_lock_irqsave(&current->sighand->siglock, flags);
	flush_signals(current);
	current->blocked = blocked_sav;
	recalc_sigpending();
	spin_unlock_irqrestore(&current->sighand->siglock, flags);

	if (err) {
		printe("Unable to contact server %s:%u (%d)\n",
		       beip_to_str(ip, ip_str), port, err);
		goto fail_socket;
	}
	*connection = sock;
	printd("connection established\n");
	return 0;

      fail_socket:
	sock_release(sock);
      fail:
	return err;
}

int
setup_server_sock(u16 * port, struct socket **connection, int maxlisten,
		  int reuse)
{
	struct socket *sock;
	struct sockaddr_in sin, result;
	int len;
	int err;

	*connection = NULL;
	err = sock_create(PF_INET, SOCK_STREAM, IPPROTO_TCP, &sock);
	if (err) {
		printe("error creating socket (%d)\n", err);
		goto fail;
	}

	if (reuse) {
		mm_segment_t fs;
		int trueint = 1;
		fs = get_fs();
		set_fs(get_ds());
		err =
		    sock_setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
				    (char *) &trueint, sizeof (int));
		set_fs(fs);
		if (err) {
			printe("error setting socket options (%d)\n", err);
			goto fail_socket;
		}
	}

	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_port = htons(*port);

	err = sock->ops->bind(sock, (struct sockaddr *) &sin, sizeof (sin));
	if (err) {
		if (port)
			printe("error binding socket to port %u (%d)\n", *port,
			       err);
		else
			printe("error binding socket (%d)\n", err);
		goto fail_socket;
	}

	err = sock->ops->getname(sock, (struct sockaddr *) &result, &len, 0);
	if (err) {
		printe("error getting socket information (%d)\n", err);
		goto fail_socket;
	}
	*port = ntohs(result.sin_port);

	err = sock->ops->listen(sock, maxlisten);
	if (err) {
		printe("error listening on socket (%d)\n", err);
		goto fail_socket;
	}

	sock->flags |= __SO_ACCEPTCON;
	*connection = sock;
	return 0;

      fail_socket:
	sock_release(sock);
      fail:
	return err;
}

int
accept_connection(struct socket *listen_sock, struct socket **new_sock)
{
	struct socket *sock;
	int err = 0;

	sock = sock_alloc();
	if (!sock) {
		printe("cannot allocate socket for new connection\n");
		err = -ENOMEM;
		goto fail;
	}

	sock->type = listen_sock->type;
	sock->ops = listen_sock->ops;

	err = listen_sock->ops->accept(listen_sock, sock, 0);
	if (err) {
		if (err != -ERESTARTSYS && err != -EINTR)
			printe("cannot accept new connection (%d)\n", err);
		goto fail_socket;
	}

	sock->sk->allocation = GFP_NOIO;
	*new_sock = sock;
	return 0;

      fail_socket:
	sock_release(sock);
      fail:
	*new_sock = NULL;
	return err;
}

/**
 * transfer_info - send or recieve one continuous buffer of information
 * @cmd: < either SEND or RECV
 * @sock: < a socket
 * @info: < > a header for a io transfer
 * @suze: < the size of the data being transfered.
 *
 * FIXME - This should be inlined
 */
int
__transfer_info(int cmd, struct socket *sock, void *info, int size,
		unsigned long timeout)
{
	struct iovec iov[1];
	iov[0].iov_base = info;
	iov[0].iov_len = size;
	return do_transfer(sock, iov, 1, iov[0].iov_len, cmd, timeout);
}

/* This function assumes that everything will work.  It would probably be
   a good idea to make sure that someone did the necessary checks before hand
*/
int
__recv_request_data(struct socket *sock, struct request *req,
		    int buf_count, unsigned long timeout)
{
	struct iovec iov[MAX_SEGMENTS];
	struct buffer_head *temp = req->bh;
	int i;
	for (i = 0; i < buf_count; i++) {
		iov[i].iov_base = temp->b_data;
		iov[i].iov_len = temp->b_size;
		temp = temp->b_reqnext;
	}
	return do_transfer(sock, iov, buf_count, buf_count * req->bh->b_size,
			   RECV, timeout);
}

/**
 * send_request_data - send the data from a request
 * @sock: < a socket
 * @req : <> a request
 *
 */
int
__send_request_data(struct socket *sock, struct request *req,
		    trans_hdr_t * header, struct iovec *iov,
		    unsigned long timeout)
{
	int num_iov = 1;
	u32 total_bytes = sizeof (trans_hdr_t);
	struct buffer_head *temp = req->bh;
	iov[0].iov_base = header;
	iov[0].iov_len = sizeof (trans_hdr_t);
	while (temp != NULL) {
		iov[num_iov].iov_base = temp->b_data;
		iov[num_iov].iov_len = temp->b_size;
		total_bytes += temp->b_size;
		num_iov++;
		temp = temp->b_reqnext;
	}
	if (num_iov == 1) {
		printe("cannot process an empty request\n");
		return -1;
	}
	return do_transfer(sock, iov, num_iov, total_bytes, SEND, timeout);
}

/**
 * trans_bh_data - transfers the data to or from a array of buffer headers
 # @cmd: < either SEND or RECV, to specify the direction of transfer
 * @sock: < a socket
 * @bh: <> an array of buffer_head pointers
 * @num: < the number of buffer heads
 * @start_offset: < the number of bytes to offset on the first buffer
 * @end_len: < the number of bytes to transmit from the last buffer
 * @total_bytes: < this is to double_check that the size matches up
 */
int
__trans_bh_data(int cmd, struct socket *sock, struct buffer_head *bh[],
		int num, int start_offset, int end_len, u32 total_bytes,
		unsigned long timeout)
{
	struct iovec iov[MAX_SEGMENTS];
	u32 check_bytes;
	int i;

	if (num <= 0) {
		printe("invalid length array : %d\n", num);
		return -1;
	}
	if (num == 1) {
		iov[0].iov_base = bh[0]->b_data + start_offset;
		iov[0].iov_len = end_len - start_offset;
		check_bytes = iov[0].iov_len;
	} else {
		iov[0].iov_base = bh[0]->b_data + start_offset;
		iov[0].iov_len = bh[0]->b_size - start_offset;
		check_bytes = iov[0].iov_len;
		for (i = 1; i < num - 1; i++) {
			iov[i].iov_base = bh[i]->b_data;
			iov[i].iov_len = bh[i]->b_size;
			check_bytes += iov[i].iov_len;
		}
		iov[i].iov_base = bh[i]->b_data;
		iov[i].iov_len = end_len;
		check_bytes += iov[i].iov_len;
	}
	if (check_bytes != total_bytes) {
		printe
		    ("trying to %s an incorrect number of bytes: want %lu have %lu\n",
		     (cmd == SEND) ? "send" : "receive",
		     (unsigned long) total_bytes, (unsigned long) check_bytes);
		return -1;
	}
	return do_transfer(sock, iov, num, total_bytes, cmd, timeout);
}

int
__transfer_be32(int cmd, struct socket *sock, uint32_t * info,
		unsigned long timeout)
{
	struct iovec iov[1];
	if (cmd == SEND) {
		uint32_t tmp = cpu_to_be32(*info);
		iov[0].iov_base = (void *) &tmp;
		iov[0].iov_len = sizeof (uint32_t);
		return do_transfer(sock, iov, 1, iov[0].iov_len, SEND, timeout);
	} else {
		int err;
		iov[0].iov_base = info;
		iov[0].iov_len = sizeof (uint32_t);
		err = do_transfer(sock, iov, 1, iov[0].iov_len, RECV, timeout);
		if (err)
			return err;
		*info = be32_to_cpu(*info);
		return 0;
	}
}

int
do_transfer(struct socket *sock, struct iovec *iov, int n, int size,
	    int dir, unsigned long timeout)
{
	struct msghdr m;
	mm_segment_t fs;
	struct timer_list timer;
	sigset_t blocked_sav;
	unsigned long flags;
	int rv, moved = 0;

	fs = get_fs();
	set_fs(get_ds());

	memset(&m, 0, sizeof (struct msghdr));

	m.msg_flags = MSG_NOSIGNAL;
	
	if (timeout) {
		init_timer(&timer);
		timer.function = send_sigusr1;
		timer.data = (unsigned long) current;
	}

	for (;;) {
		m.msg_iov = iov;
		m.msg_iovlen = n;

		if (timeout) {
			spin_lock_irqsave(&current->sighand->siglock, flags);
			blocked_sav = current->blocked;
			sigdelsetmask(&current->blocked, sigmask(SIGUSR1));
			recalc_sigpending();
			spin_unlock_irqrestore(&current->sighand->siglock,
					       flags);

			mod_timer(&timer, jiffies + (timeout * HZ));
		}

		if (dir)
			rv = sock_sendmsg(sock, &m, size - moved);
		else
			rv = sock_recvmsg(sock, &m, size - moved, 0);

		if (timeout) {
			del_timer(&timer);

			spin_lock_irqsave(&current->sighand->siglock, flags);
			flush_signals(current);
			current->blocked = blocked_sav;
			recalc_sigpending();
			spin_unlock_irqrestore(&current->sighand->siglock,
					       flags);
		}

		if (rv <= 0)
			goto out_err;
		moved += rv;

		if (moved >= size)
			break;

		/* adjust iov's for next transfer */

		while (iov->iov_len == 0) {
			n--;
			iov++;
		}
	}

	set_fs(fs);
	return 0;

      out_err:
	set_fs(fs);
	if (!rv)
		rv = -1;
	return rv;
}
